#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.data_manipulation.remote_objects.versions.v_0_fast_optimized_numba.remote_objects import *
from cengal.data_manipulation.remote_objects.versions.v_0_fast.remote_objects import RemoteObjectsManager as FastRemoteObjectsManager
from cengal.data_manipulation.remote_objects.versions.v_0.remote_objects import RemoteObjectsManager as UniversalRemoteObjectsManager
from cengal.data_manipulation.remote_objects.versions.v_0_optimized.remote_objects import RemoteObjectsManager as UniversalOptimizedRemoteObjectsManager
from cengal.code_flow_control.smart_values import ResultExistence
from cengal.code_flow_control.gc import DisableGC, EnableGC
from cengal.performance_test_lib import MeasureTime, measure_func_isolated_performance
from cengal.data_manipulation.serialization import test_data_factory, TestDataType, best_serializer, \
    SerializerFeatures, Tags, DataFormats, get_most_efficient_serializers, SUPPORTED_TYPES, Serializers, \
    Serializer
from cengal.introspection.inspect import cen
# from cengal.code_inspection.line_profiling import set_profiler, profiler_result
from collections import deque
from dataclasses import dataclass
from pickle import loads as pickle_loads, dumps as pickle_dumps
from marshal import loads as marshal_loads, dumps as marshal_dumps
from pprint import pprint
from unittest import TestCase, main, SkipTest
from typing import Any, Dict, Optional, Callable, Set, Type, Tuple, List, FrozenSet, Deque, Generator


'''
>>> "DataClass": func()
         It was used 0.375 seconds to make 1842 iterations. Performance: 4912.0 iterations/seconds
         Isolated run time: 6.055354606360197e-05 seconds; Isolated performance: 16514.309483207762 iterations/seconds

>>> "DataClass": func_marshal()
         It was used 1.75 seconds to make 5318 iterations. Performance: 3038.857177734375 iterations/seconds
         Isolated run time: 4.380743484944105e-05 seconds; Isolated performance: 22827.175419807973 iterations/seconds

>>> "TestDataType.small": func()
         It was used 0.0625 seconds to make 97 iterations. Performance: 1552.0 iterations/seconds
         Isolated run time: 0.00015537731815129519 seconds; Isolated performance: 6435.945811770753 iterations/seconds

>>> "TestDataType.small": func_marshal()
         It was used 0.0625 seconds to make 223 iterations. Performance: 3568.0 iterations/seconds
         Isolated run time: 0.0001032799482345581 seconds; Isolated performance: 9682.421584186986 iterations/seconds

>>> "TestDataType.deep_small": func()
         It was used 0.4375 seconds to make 2 iterations. Performance: 4.5714287757873535 iterations/seconds
         Isolated run time: 0.21644665556959808 seconds; Isolated performance: 4.620076006110668 iterations/seconds

>>> "TestDataType.deep_small": func_marshal()
         It was used 0.5 seconds to make 2 iterations. Performance: 4.0 iterations/seconds
         Isolated run time: 0.15614438650663942 seconds; Isolated performance: 6.404328854675022 iterations/seconds

>>> "TestDataType.large": func()
         It was used 0.0625 seconds to make 63 iterations. Performance: 1008.0 iterations/seconds
         Isolated run time: 0.00015381828416138887 seconds; Isolated performance: 6501.177707526514 iterations/seconds

>>> "TestDataType.large": func_marshal()
         It was used 0.125 seconds to make 96 iterations. Performance: 768.0 iterations/seconds
         Isolated run time: 0.000106387073174119 seconds; Isolated performance: 9399.638228258656 iterations/seconds

>>> "TestDataType.deep_large": func()
         It was used 0.6875 seconds to make 2 iterations. Performance: 2.909090995788574 iterations/seconds
         Isolated run time: 0.28062836138997227 seconds; Isolated performance: 3.5634317039337318 iterations/seconds

>>> "TestDataType.deep_large": func_marshal()
         It was used 0.25 seconds to make 2 iterations. Performance: 8.0 iterations/seconds
         Isolated run time: 0.12075315089896321 seconds; Isolated performance: 8.28135740190102 iterations/seconds

=== Serializers on DataClass: ==========================================================================
[(<Serializers.pickle: 16>, 84079.85779726713, 113),
 (<Serializers.cloudpickle: 30>, 1526.005642177909, 3094)]

=== Serializers on TestDataType: =======================================================================
        == TestDataType.small: ==
                == current_platform: ==
[(<Serializers.msgspec_messagepack: 37>, 257731.54286057188, 77),
 (<Serializers.cbor: 8>, 243044.86297145122, 81),
 (<Serializers.marshal: 10>, 228973.3331200853, 129),
 (<Serializers.msgspec_json: 36>, 226468.08837331927, 151),
 (<Serializers.pickle: 16>, 226289.10937829295, 116),
 (<Serializers.orjson: 3>, 210043.39280125196, 151),
 (<Serializers.cloudpickle: 30>, 128098.999239453, 116),
 (<Serializers.msgpack: 7>, 123695.84977823858, 77),
 (<Serializers.ujson: 2>, 111154.83626858526, 151),
 (<Serializers.cbor2: 9>, 55260.28236353694, 81),
 (<Serializers.json: 0>, 54488.06576677154, 180),
 (<Serializers.simplejson: 1>, 37297.109525815096, 180),
 (<Serializers.msgspec_toml: 39>, 2870.8551811309517, 236),
 (<Serializers.msgspec_yaml: 38>, 5.70717582925687, 176)]

                == multi_platform: ==
[(<Serializers.msgspec_messagepack: 37>, 256354.73892801718, 77),
 (<Serializers.cbor: 8>, 242564.44221048767, 81),
 (<Serializers.msgspec_json: 36>, 225777.60058876098, 151),
 (<Serializers.orjson: 3>, 210279.9165728274, 151),
 (<Serializers.msgpack: 7>, 126290.99478071659, 77),
 (<Serializers.ujson: 2>, 112623.86217566309, 151),
 (<Serializers.cbor2: 9>, 55535.73704695036, 81),
 (<Serializers.json: 0>, 53815.913041843916, 180),
 (<Serializers.simplejson: 1>, 36832.36896109632, 180),
 (<Serializers.msgspec_toml: 39>, 2860.5225212084883, 236),
 (<Serializers.msgspec_yaml: 38>, 1865.1935922838156, 176)]

        == TestDataType.deep_small: ==
                == current_platform: ==
[(<Serializers.marshal: 10>, 34535.87131168962, 5389),
 (<Serializers.pickle: 16>, 33390.9987133289, 2220),
 (<Serializers.cloudpickle: 30>, 28798.032036797395, 2220),
 (<Serializers.msgspec_messagepack: 37>, 3144.473588417121, 20529),
 (<Serializers.orjson: 3>, 2168.3754029132183, 41007),
 (<Serializers.msgspec_json: 36>, 1774.1164134899966, 41007),
 (<Serializers.cbor: 8>, 1144.593445973379, 20568),
 (<Serializers.msgpack: 7>, 1120.6912215157654, 20529),
 (<Serializers.ujson: 2>, 791.5599158746937, 41007),
 (<Serializers.cbor2: 9>, 516.688571982815, 20568),
 (<Serializers.json: 0>, 317.2189294063965, 61194),
 (<Serializers.simplejson: 1>, 222.95206437015122, 61194),
 (<Serializers.msgspec_toml: 39>, 3.254262227964867, 141522),
 (<Serializers.msgspec_yaml: 38>, 1.0197201679345653, 161098)]

                == multi_platform: ==
[(<Serializers.msgspec_messagepack: 37>, 3134.421203006586, 20529),
 (<Serializers.orjson: 3>, 2171.4334734959234, 41007),
 (<Serializers.msgspec_json: 36>, 1870.9012938502635, 41007),
 (<Serializers.cbor: 8>, 1146.3791282140733, 20568),
 (<Serializers.msgpack: 7>, 1118.7989337595654, 20529),
 (<Serializers.ujson: 2>, 805.304983882059, 41007),
 (<Serializers.cbor2: 9>, 519.821482140335, 20568),
 (<Serializers.json: 0>, 372.63981276767424, 61194),
 (<Serializers.simplejson: 1>, 235.25868771211785, 61194),
 (<Serializers.msgspec_toml: 39>, 10.792663591483793, 141522),
 (<Serializers.msgspec_yaml: 38>, 1.5124533193200795, 161098)]

        == TestDataType.large: ==
                == current_platform: ==
[(<Serializers.pickle: 16>, 198703.09026139256, 1510),
 (<Serializers.marshal: 10>, 185844.84524350404, 1523),
 (<Serializers.msgspec_messagepack: 37>, 184892.79993112208, 2065),
 (<Serializers.cbor: 8>, 181993.99546600564, 2069),
 (<Serializers.orjson: 3>, 131864.76608025545, 2129),
 (<Serializers.msgspec_json: 36>, 131650.54242275626, 2129),
 (<Serializers.cloudpickle: 30>, 119035.1647243047, 1510),
 (<Serializers.msgpack: 7>, 104186.08810401708, 2065),
 (<Serializers.ujson: 2>, 69493.91694645125, 2129),
 (<Serializers.cbor2: 9>, 47493.36299802063, 2069),
 (<Serializers.json: 0>, 38580.2650426003, 2158),
 (<Serializers.simplejson: 1>, 29288.194592382966, 2158),
 (<Serializers.msgspec_yaml: 38>, 1768.7227299073936, 2172),
 (<Serializers.msgspec_toml: 39>, 1755.31904382405, 2214)]

                == multi_platform: ==
[(<Serializers.cbor: 8>, 190934.10815977238, 2069),
 (<Serializers.msgspec_messagepack: 37>, 176030.46419935243, 2065),
 (<Serializers.msgspec_json: 36>, 125230.4840435613, 2129),
 (<Serializers.orjson: 3>, 124770.27847659995, 2129),
 (<Serializers.msgpack: 7>, 109767.1052954406, 2065),
 (<Serializers.ujson: 2>, 65987.59048972536, 2129),
 (<Serializers.cbor2: 9>, 49679.794755531904, 2069),
 (<Serializers.json: 0>, 36743.98186313511, 2158),
 (<Serializers.simplejson: 1>, 27627.030930288656, 2158),
 (<Serializers.msgspec_yaml: 38>, 1739.2710742439529, 2172),
 (<Serializers.msgspec_toml: 39>, 1683.5819118815534, 2214)]

        == TestDataType.deep_large: ==
                == current_platform: ==
[(<Serializers.marshal: 10>, 33605.10219315061, 6783),
 (<Serializers.pickle: 16>, 31551.06276101449, 3614),
 (<Serializers.cloudpickle: 30>, 28131.53012454601, 3614),
 (<Serializers.msgspec_messagepack: 37>, 3021.7945839460226, 33245),
 (<Serializers.orjson: 3>, 2130.981857242975, 53659),
 (<Serializers.msgspec_json: 36>, 1848.7170993644584, 53659),
 (<Serializers.cbor: 8>, 1104.750371714677, 33284),
 (<Serializers.msgpack: 7>, 1099.680116364709, 33245),
 (<Serializers.ujson: 2>, 788.3222605179876, 53659),
 (<Serializers.cbor2: 9>, 500.9584529072141, 33284),
 (<Serializers.json: 0>, 367.9899350773419, 73846),
 (<Serializers.simplejson: 1>, 260.480108583734, 73846),
 (<Serializers.msgspec_toml: 39>, 10.986507566842041, 154174),
 (<Serializers.msgspec_yaml: 38>, 3.073031855751556, 173930)]

                == multi_platform: ==
[(<Serializers.msgspec_messagepack: 37>, 3129.803900078701, 33245),
 (<Serializers.orjson: 3>, 2071.6953354664643, 53659),
 (<Serializers.msgspec_json: 36>, 1872.0045991950972, 53659),
 (<Serializers.cbor: 8>, 1144.0966152329834, 33284),
 (<Serializers.msgpack: 7>, 1098.575114462423, 33245),
 (<Serializers.ujson: 2>, 802.7893724457039, 53659),
 (<Serializers.cbor2: 9>, 515.0703656235943, 33284),
 (<Serializers.json: 0>, 361.8401938939023, 73846),
 (<Serializers.simplejson: 1>, 259.5755346231525, 73846),
 (<Serializers.msgspec_toml: 39>, 10.768618751382412, 154174),
 (<Serializers.msgspec_yaml: 38>, 3.9636315642372315, 173930)]

=== RemoteObjectsManager: ==============================================================================
>>> "0, Serializers.marshal, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 256 iterations. Performance: 4096.0 iterations/seconds
         Isolated run time: 0.0002806582488119602 seconds; Isolated performance: 3563.0522325035795 iterations/seconds

>>> "0, Serializers.marshal, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 798 iterations. Performance: 12768.0 iterations/seconds
         Isolated run time: 9.579467587172985e-05 seconds; Isolated performance: 10438.993512948582 iterations/seconds

>>> "0, Serializers.marshal, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 200 iterations. Performance: 1600.0 iterations/seconds
         Isolated run time: 0.0003355335211381316 seconds; Isolated performance: 2980.328155017103 iterations/seconds

>>> "0, Serializers.marshal, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 216 iterations. Performance: 3456.0 iterations/seconds
         Isolated run time: 0.00012660003267228603 seconds; Isolated performance: 7898.892116312302 iterations/seconds

>>> "0, Serializers.pickle, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 377 iterations. Performance: 6032.0 iterations/seconds
         Isolated run time: 0.00017896434292197227 seconds; Isolated performance: 5587.705258337388 iterations/seconds

>>> "0, Serializers.pickle, universal=False, optimized=True": benchmark_single_item()
         It was used 0.1875 seconds to make 1008 iterations. Performance: 5376.0 iterations/seconds
         Isolated run time: 7.234816439449787e-05 seconds; Isolated performance: 13822.050750966264 iterations/seconds

>>> "0, Serializers.pickle, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 125 iterations. Performance: 1000.0 iterations/seconds
         Isolated run time: 0.00019822618924081326 seconds; Isolated performance: 5044.742088973719 iterations/seconds

>>> "0, Serializers.pickle, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 408 iterations. Performance: 6528.0 iterations/seconds
         Isolated run time: 8.212903048843145e-05 seconds; Isolated performance: 12175.962556149474 iterations/seconds

>>> "0, Serializers.cloudpickle, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 50 iterations. Performance: 800.0 iterations/seconds
         Isolated run time: 0.0009289330337196589 seconds; Isolated performance: 1076.5038637885154 iterations/seconds

>>> "0, Serializers.cloudpickle, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 98 iterations. Performance: 784.0 iterations/seconds
         Isolated run time: 0.0007491244468837976 seconds; Isolated performance: 1334.891691440311 iterations/seconds

>>> "0, Serializers.cloudpickle, universal=True, optimized=False": benchmark_single_item()
         It was used 0.1875 seconds to make 88 iterations. Performance: 469.3333435058594 iterations/seconds
         Isolated run time: 0.0009302576072514057 seconds; Isolated performance: 1074.9710533995624 iterations/seconds

>>> "0, Serializers.cloudpickle, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 92 iterations. Performance: 736.0 iterations/seconds
         Isolated run time: 0.000772637315094471 seconds; Isolated performance: 1294.2683202890987 iterations/seconds

>>> "0, Serializers.msgspec_messagepack, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 263 iterations. Performance: 2104.0 iterations/seconds
         Isolated run time: 0.00028666818980127573 seconds; Isolated performance: 3488.353558492906 iterations/seconds

>>> "0, Serializers.msgspec_messagepack, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 697 iterations. Performance: 11152.0 iterations/seconds
         Isolated run time: 0.00010353734251111746 seconds; Isolated performance: 9658.351042693836 iterations/seconds

>>> "0, Serializers.msgspec_messagepack, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 201 iterations. Performance: 1608.0 iterations/seconds
         Isolated run time: 0.0003793839132413268 seconds; Isolated performance: 2635.8524046429407 iterations/seconds

>>> "0, Serializers.msgspec_messagepack, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 335 iterations. Performance: 5360.0 iterations/seconds
         Isolated run time: 0.0001441108761355281 seconds; Isolated performance: 6939.101522494088 iterations/seconds

>>> "0, Serializers.orjson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 211 iterations. Performance: 3376.0 iterations/seconds
         Isolated run time: 0.0003095826832577586 seconds; Isolated performance: 3230.154831261669 iterations/seconds

>>> "0, Serializers.orjson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 658 iterations. Performance: 5264.0 iterations/seconds
         Isolated run time: 0.00011941057164222002 seconds; Isolated performance: 8374.46790721526 iterations/seconds

>>> "0, Serializers.orjson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 132 iterations. Performance: 2112.0 iterations/seconds
         Isolated run time: 0.00040324265137314796 seconds; Isolated performance: 2479.8964013224677 iterations/seconds

>>> "0, Serializers.orjson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 207 iterations. Performance: 3312.0 iterations/seconds
         Isolated run time: 0.00016040436457842588 seconds; Isolated performance: 6234.24432762909 iterations/seconds

>>> "0, Serializers.msgspec_json, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 130 iterations. Performance: 2080.0 iterations/seconds
         Isolated run time: 0.00029032560996711254 seconds; Isolated performance: 3444.408504345441 iterations/seconds

>>> "0, Serializers.msgspec_json, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 442 iterations. Performance: 7072.0 iterations/seconds
         Isolated run time: 0.00011189084034413099 seconds; Isolated performance: 8937.282059232053 iterations/seconds

>>> "0, Serializers.msgspec_json, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 212 iterations. Performance: 1696.0 iterations/seconds
         Isolated run time: 0.00037433707620948553 seconds; Isolated performance: 2671.389139771938 iterations/seconds

>>> "0, Serializers.msgspec_json, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 563 iterations. Performance: 9008.0 iterations/seconds
         Isolated run time: 0.00014174007810652256 seconds; Isolated performance: 7055.167552881307 iterations/seconds

>>> "0, Serializers.cbor, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 242 iterations. Performance: 3872.0 iterations/seconds
         Isolated run time: 0.00027696089819073677 seconds; Isolated performance: 3610.6179844611943 iterations/seconds

>>> "0, Serializers.cbor, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 696 iterations. Performance: 5568.0 iterations/seconds
         Isolated run time: 9.596534073352814e-05 seconds; Isolated performance: 10420.428796024922 iterations/seconds

>>> "0, Serializers.cbor, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 209 iterations. Performance: 1672.0 iterations/seconds
         Isolated run time: 0.00035776791628450155 seconds; Isolated performance: 2795.108097968146 iterations/seconds

>>> "0, Serializers.cbor, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 373 iterations. Performance: 2984.0 iterations/seconds
         Isolated run time: 0.00013779383152723312 seconds; Isolated performance: 7257.218911155419 iterations/seconds

>>> "0, Serializers.msgpack, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 155 iterations. Performance: 2480.0 iterations/seconds
         Isolated run time: 0.0003367925528436899 seconds; Isolated performance: 2969.186793343717 iterations/seconds

>>> "0, Serializers.msgpack, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 492 iterations. Performance: 7872.0 iterations/seconds
         Isolated run time: 0.00013641652185469866 seconds; Isolated performance: 7330.490371724402 iterations/seconds

>>> "0, Serializers.msgpack, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 192 iterations. Performance: 1536.0 iterations/seconds
         Isolated run time: 0.00042097328696399927 seconds; Isolated performance: 2375.447637573065 iterations/seconds

>>> "0, Serializers.msgpack, universal=True, optimized=True": benchmark_single_item()
         It was used 0.1875 seconds to make 288 iterations. Performance: 1536.0 iterations/seconds
         Isolated run time: 0.0001829200191423297 seconds; Isolated performance: 5466.870191074614 iterations/seconds

>>> "0, Serializers.ujson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 97 iterations. Performance: 1552.0 iterations/seconds
         Isolated run time: 0.00034592265728861094 seconds; Isolated performance: 2890.8196064349663 iterations/seconds

>>> "0, Serializers.ujson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 202 iterations. Performance: 3232.0 iterations/seconds
         Isolated run time: 0.00013985077384859324 seconds; Isolated performance: 7150.478845992164 iterations/seconds

>>> "0, Serializers.ujson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 170 iterations. Performance: 2720.0 iterations/seconds
         Isolated run time: 0.00041919376235455275 seconds; Isolated performance: 2385.5316796298202 iterations/seconds

>>> "0, Serializers.ujson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 440 iterations. Performance: 3520.0 iterations/seconds
         Isolated run time: 0.00017518852837383747 seconds; Isolated performance: 5708.136310535612 iterations/seconds

>>> "0, Serializers.cbor2, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 124 iterations. Performance: 1984.0 iterations/seconds
         Isolated run time: 0.0004588782321661711 seconds; Isolated performance: 2179.227363388803 iterations/seconds

>>> "0, Serializers.cbor2, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 311 iterations. Performance: 4976.0 iterations/seconds
         Isolated run time: 0.00023057474754750729 seconds; Isolated performance: 4336.988376378734 iterations/seconds

>>> "0, Serializers.cbor2, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 124 iterations. Performance: 1984.0 iterations/seconds
         Isolated run time: 0.0005508470349013805 seconds; Isolated performance: 1815.386008529632 iterations/seconds

>>> "0, Serializers.cbor2, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 281 iterations. Performance: 2248.0 iterations/seconds
         Isolated run time: 0.0002779026981443167 seconds; Isolated performance: 3598.381759793831 iterations/seconds

>>> "0, Serializers.json, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 115 iterations. Performance: 1840.0 iterations/seconds
         Isolated run time: 0.00043530913535505533 seconds; Isolated performance: 2297.218042953223 iterations/seconds

>>> "0, Serializers.json, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 377 iterations. Performance: 3016.0 iterations/seconds
         Isolated run time: 0.00020241155289113522 seconds; Isolated performance: 4940.429465198752 iterations/seconds

>>> "0, Serializers.json, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 156 iterations. Performance: 1248.0 iterations/seconds
         Isolated run time: 0.000503272982314229 seconds; Isolated performance: 1986.9932127125971 iterations/seconds

>>> "0, Serializers.json, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 307 iterations. Performance: 2456.0 iterations/seconds
         Isolated run time: 0.00023740611504763365 seconds; Isolated performance: 4212.191416380989 iterations/seconds

>>> "0, Serializers.simplejson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.1875 seconds to make 128 iterations. Performance: 682.6666870117188 iterations/seconds
         Isolated run time: 0.0004994800547137856 seconds; Isolated performance: 2002.0819461410217 iterations/seconds

>>> "0, Serializers.simplejson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.1875 seconds to make 133 iterations. Performance: 709.3333129882812 iterations/seconds
         Isolated run time: 0.00026223529130220413 seconds; Isolated performance: 3813.369264808771 iterations/seconds

>>> "0, Serializers.simplejson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 50 iterations. Performance: 800.0 iterations/seconds
         Isolated run time: 0.0006034626858308911 seconds; Isolated performance: 1657.1032865489067 iterations/seconds

>>> "0, Serializers.simplejson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 237 iterations. Performance: 3792.0 iterations/seconds
         Isolated run time: 0.0003024700563400984 seconds; Isolated performance: 3306.112387123691 iterations/seconds

>>> "0, Serializers.msgspec_toml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 23 iterations. Performance: 184.0 iterations/seconds
         Isolated run time: 0.003502200823277235 seconds; Isolated performance: 285.53473957105507 iterations/seconds

>>> "0, Serializers.msgspec_toml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 24 iterations. Performance: 384.0 iterations/seconds
         Isolated run time: 0.00302437343634665 seconds; Isolated performance: 330.64699880712124 iterations/seconds

>>> "0, Serializers.msgspec_toml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 24 iterations. Performance: 384.0 iterations/seconds
         Isolated run time: 0.0033045709133148193 seconds; Isolated performance: 302.611148688257 iterations/seconds

>>> "0, Serializers.msgspec_toml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 26 iterations. Performance: 208.0 iterations/seconds
         Isolated run time: 0.00302398344501853 seconds; Isolated performance: 330.6896410584921 iterations/seconds

>>> "0, Serializers.msgspec_yaml, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 16 iterations. Performance: 128.0 iterations/seconds
         Isolated run time: 0.0025470209075137973 seconds; Isolated performance: 392.61554432080493 iterations/seconds

>>> "0, Serializers.msgspec_yaml, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 39 iterations. Performance: 624.0 iterations/seconds
         Isolated run time: 0.0019713385263457894 seconds; Isolated performance: 507.2695463694253 iterations/seconds

>>> "0, Serializers.msgspec_yaml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 34 iterations. Performance: 544.0 iterations/seconds
         Isolated run time: 0.0024683665251359344 seconds; Isolated performance: 405.12622003935553 iterations/seconds

>>> "0, Serializers.msgspec_yaml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 44 iterations. Performance: 704.0 iterations/seconds
         Isolated run time: 0.0020248856162652373 seconds; Isolated performance: 493.85505628926904 iterations/seconds

>>> "1, Serializers.marshal, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 91 iterations. Performance: 1456.0 iterations/seconds
         Isolated run time: 0.0008453736081719398 seconds; Isolated performance: 1182.908941482605 iterations/seconds

>>> "1, Serializers.marshal, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 128 iterations. Performance: 2048.0 iterations/seconds
         Isolated run time: 0.0002721799537539482 seconds; Isolated performance: 3674.0398629944807 iterations/seconds

>>> "1, Serializers.marshal, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 40 iterations. Performance: 320.0 iterations/seconds
         Isolated run time: 0.0009854146046563983 seconds; Isolated performance: 1014.8012778323774 iterations/seconds

>>> "1, Serializers.marshal, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 129 iterations. Performance: 2064.0 iterations/seconds
         Isolated run time: 0.00032548257149755955 seconds; Isolated performance: 3072.3611264313054 iterations/seconds

>>> "1, Serializers.pickle, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 81 iterations. Performance: 648.0 iterations/seconds
         Isolated run time: 0.0009101399919018149 seconds; Isolated performance: 1098.7320729752958 iterations/seconds

>>> "1, Serializers.pickle, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 266 iterations. Performance: 2128.0 iterations/seconds
         Isolated run time: 0.00029449432622641325 seconds; Isolated performance: 3395.6511584239474 iterations/seconds

>>> "1, Serializers.pickle, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 79 iterations. Performance: 632.0 iterations/seconds
         Isolated run time: 0.001000366173684597 seconds; Isolated performance: 999.6339603494905 iterations/seconds

>>> "1, Serializers.pickle, universal=True, optimized=True": benchmark_single_item()
         It was used 0.25 seconds to make 205 iterations. Performance: 820.0 iterations/seconds
         Isolated run time: 0.000346930930390954 seconds; Isolated performance: 2882.4181195752913 iterations/seconds

>>> "1, Serializers.cloudpickle, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 34 iterations. Performance: 544.0 iterations/seconds
         Isolated run time: 0.001132475328631699 seconds; Isolated performance: 883.0214440152432 iterations/seconds

>>> "1, Serializers.cloudpickle, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 165 iterations. Performance: 2640.0 iterations/seconds
         Isolated run time: 0.0004302131710574031 seconds; Isolated performance: 2324.429067436828 iterations/seconds

>>> "1, Serializers.cloudpickle, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 60 iterations. Performance: 480.0 iterations/seconds
         Isolated run time: 0.0011957543902099133 seconds; Isolated performance: 836.2921417536684 iterations/seconds

>>> "1, Serializers.cloudpickle, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 176 iterations. Performance: 2816.0 iterations/seconds
         Isolated run time: 0.000480318209156394 seconds; Isolated performance: 2081.953132187822 iterations/seconds

>>> "1, Serializers.msgspec_messagepack, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 93 iterations. Performance: 1488.0 iterations/seconds
         Isolated run time: 0.0008552558720111847 seconds; Isolated performance: 1169.2407298513378 iterations/seconds

>>> "1, Serializers.msgspec_messagepack, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 199 iterations. Performance: 1592.0 iterations/seconds
         Isolated run time: 0.0002988508203998208 seconds; Isolated performance: 3346.1510952592976 iterations/seconds

>>> "1, Serializers.msgspec_messagepack, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 38 iterations. Performance: 304.0 iterations/seconds
         Isolated run time: 0.0011214744299650192 seconds; Isolated performance: 891.6832816519872 iterations/seconds

>>> "1, Serializers.msgspec_messagepack, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 100 iterations. Performance: 800.0 iterations/seconds
         Isolated run time: 0.0003879438154399395 seconds; Isolated performance: 2577.6928519042663 iterations/seconds

>>> "1, Serializers.orjson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 68 iterations. Performance: 544.0 iterations/seconds
         Isolated run time: 0.0009632299188524485 seconds; Isolated performance: 1038.173732385055 iterations/seconds

>>> "1, Serializers.orjson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.1875 seconds to make 235 iterations. Performance: 1253.3333740234375 iterations/seconds
         Isolated run time: 0.0003308698069304228 seconds; Isolated performance: 3022.336819661172 iterations/seconds

>>> "1, Serializers.orjson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 60 iterations. Performance: 960.0 iterations/seconds
         Isolated run time: 0.0011929599568247795 seconds; Isolated performance: 838.2511032990848 iterations/seconds

>>> "1, Serializers.orjson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 101 iterations. Performance: 808.0 iterations/seconds
         Isolated run time: 0.00044839445035904646 seconds; Isolated performance: 2230.1792522170203 iterations/seconds

>>> "1, Serializers.msgspec_json, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 57 iterations. Performance: 912.0 iterations/seconds
         Isolated run time: 0.0008861626265570521 seconds; Isolated performance: 1128.4610409324443 iterations/seconds

>>> "1, Serializers.msgspec_json, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 250 iterations. Performance: 4000.0 iterations/seconds
         Isolated run time: 0.000300672953017056 seconds; Isolated performance: 3325.872812854151 iterations/seconds

>>> "1, Serializers.msgspec_json, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 76 iterations. Performance: 1216.0 iterations/seconds
         Isolated run time: 0.0011155385291203856 seconds; Isolated performance: 896.4280245779686 iterations/seconds

>>> "1, Serializers.msgspec_json, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 214 iterations. Performance: 3424.0 iterations/seconds
         Isolated run time: 0.00038392271380871534 seconds; Isolated performance: 2604.690902706625 iterations/seconds

>>> "1, Serializers.cbor, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 101 iterations. Performance: 808.0 iterations/seconds
         Isolated run time: 0.0008222202304750681 seconds; Isolated performance: 1216.219162379662 iterations/seconds

>>> "1, Serializers.cbor, universal=False, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 313 iterations. Performance: 2504.0 iterations/seconds
         Isolated run time: 0.00027243385557085276 seconds; Isolated performance: 3670.6157459931655 iterations/seconds

>>> "1, Serializers.cbor, universal=True, optimized=False": benchmark_single_item()
         It was used 0.25 seconds to make 43 iterations. Performance: 172.0 iterations/seconds
         Isolated run time: 0.0011006708955392241 seconds; Isolated performance: 908.5367879288705 iterations/seconds

>>> "1, Serializers.cbor, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 93 iterations. Performance: 1488.0 iterations/seconds
         Isolated run time: 0.0003714247141033411 seconds; Isolated performance: 2692.3356524998794 iterations/seconds

>>> "1, Serializers.msgpack, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 48 iterations. Performance: 768.0 iterations/seconds
         Isolated run time: 0.0010566673008725047 seconds; Isolated performance: 946.3716717402784 iterations/seconds

>>> "1, Serializers.msgpack, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 202 iterations. Performance: 3232.0 iterations/seconds
         Isolated run time: 0.0003928329097107053 seconds; Isolated performance: 2545.611569907501 iterations/seconds

>>> "1, Serializers.msgpack, universal=True, optimized=False": benchmark_single_item()
         It was used 0.1875 seconds to make 53 iterations. Performance: 282.6666564941406 iterations/seconds
         Isolated run time: 0.0013008699752390385 seconds; Isolated performance: 768.7163352480691 iterations/seconds

>>> "1, Serializers.msgpack, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 48 iterations. Performance: 768.0 iterations/seconds
         Isolated run time: 0.0005075300578027964 seconds; Isolated performance: 1970.3266528276351 iterations/seconds

>>> "1, Serializers.ujson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 28 iterations. Performance: 448.0 iterations/seconds
         Isolated run time: 0.0010799691081047058 seconds; Isolated performance: 925.9524115045774 iterations/seconds

>>> "1, Serializers.ujson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 123 iterations. Performance: 1968.0 iterations/seconds
         Isolated run time: 0.00039702479261904955 seconds; Isolated performance: 2518.734392890957 iterations/seconds

>>> "1, Serializers.ujson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 54 iterations. Performance: 864.0 iterations/seconds
         Isolated run time: 0.001291462336666882 seconds; Isolated performance: 774.3160381903871 iterations/seconds

>>> "1, Serializers.ujson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 174 iterations. Performance: 1392.0 iterations/seconds
         Isolated run time: 0.0004812729312106967 seconds; Isolated performance: 2077.8230711716665 iterations/seconds

>>> "1, Serializers.cbor2, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 56 iterations. Performance: 448.0 iterations/seconds
         Isolated run time: 0.0014327315147966146 seconds; Isolated performance: 697.9674765805346 iterations/seconds

>>> "1, Serializers.cbor2, universal=False, optimized=True": benchmark_single_item()
         It was used 0.1875 seconds to make 121 iterations. Performance: 645.3333129882812 iterations/seconds
         Isolated run time: 0.0006749002495780587 seconds; Isolated performance: 1481.7004448067557 iterations/seconds

>>> "1, Serializers.cbor2, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 47 iterations. Performance: 376.0 iterations/seconds
         Isolated run time: 0.0017391329165548086 seconds; Isolated performance: 574.9991794652374 iterations/seconds

>>> "1, Serializers.cbor2, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 54 iterations. Performance: 432.0 iterations/seconds
         Isolated run time: 0.0008115472737699747 seconds; Isolated performance: 1232.2141079404826 iterations/seconds

>>> "1, Serializers.json, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 31 iterations. Performance: 248.0 iterations/seconds
         Isolated run time: 0.0013661317061632872 seconds; Isolated performance: 731.9938447285219 iterations/seconds

>>> "1, Serializers.json, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 51 iterations. Performance: 816.0 iterations/seconds
         Isolated run time: 0.0006226717960089445 seconds; Isolated performance: 1605.9824877400345 iterations/seconds

>>> "1, Serializers.json, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 43 iterations. Performance: 344.0 iterations/seconds
         Isolated run time: 0.0016000053146854043 seconds; Isolated performance: 624.9979239579098 iterations/seconds

>>> "1, Serializers.json, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 70 iterations. Performance: 1120.0 iterations/seconds
         Isolated run time: 0.0007193082710728049 seconds; Isolated performance: 1390.224525721858 iterations/seconds

>>> "1, Serializers.simplejson, universal=False, optimized=False": benchmark_single_item()
         It was used 0.0 seconds to make 22 iterations. Performance: 0.0 iterations/seconds
         Isolated run time: 0.0015618717297911644 seconds; Isolated performance: 640.2574429935476 iterations/seconds

>>> "1, Serializers.simplejson, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 109 iterations. Performance: 1744.0 iterations/seconds
         Isolated run time: 0.000766547629609704 seconds; Isolated performance: 1304.550378049125 iterations/seconds

>>> "1, Serializers.simplejson, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 49 iterations. Performance: 784.0 iterations/seconds
         Isolated run time: 0.0017996785463765264 seconds; Isolated performance: 555.6547873582205 iterations/seconds

>>> "1, Serializers.simplejson, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 93 iterations. Performance: 1488.0 iterations/seconds
         Isolated run time: 0.0008633049437776208 seconds; Isolated performance: 1158.339248729694 iterations/seconds

>>> "1, Serializers.msgspec_toml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.0625 seconds to make 10 iterations. Performance: 160.0 iterations/seconds
         Isolated run time: 0.010459218523465097 seconds; Isolated performance: 95.60943752695435 iterations/seconds

>>> "1, Serializers.msgspec_toml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.125 seconds to make 11 iterations. Performance: 88.0 iterations/seconds
         Isolated run time: 0.009011557092890143 seconds; Isolated performance: 110.96861393565058 iterations/seconds

>>> "1, Serializers.msgspec_toml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 9 iterations. Performance: 72.0 iterations/seconds
         Isolated run time: 0.01117266295477748 seconds; Isolated performance: 89.50417676140455 iterations/seconds

>>> "1, Serializers.msgspec_toml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.25 seconds to make 10 iterations. Performance: 40.0 iterations/seconds
         Isolated run time: 0.009097612579353154 seconds; Isolated performance: 109.9189475565798 iterations/seconds

>>> "1, Serializers.msgspec_yaml, universal=False, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 7 iterations. Performance: 56.0 iterations/seconds
         Isolated run time: 0.016271531814709306 seconds; Isolated performance: 61.457028839534935 iterations/seconds

>>> "1, Serializers.msgspec_yaml, universal=False, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 7 iterations. Performance: 112.0 iterations/seconds
         Isolated run time: 0.006149194668978453 seconds; Isolated performance: 162.62292118426737 iterations/seconds

>>> "1, Serializers.msgspec_yaml, universal=True, optimized=False": benchmark_single_item()
         It was used 0.125 seconds to make 7 iterations. Performance: 56.0 iterations/seconds
         Isolated run time: 0.007870391476899385 seconds; Isolated performance: 127.05848278768967 iterations/seconds

>>> "1, Serializers.msgspec_yaml, universal=True, optimized=True": benchmark_single_item()
         It was used 0.0625 seconds to make 9 iterations. Performance: 144.0 iterations/seconds
         Isolated run time: 0.005934227025136352 seconds; Isolated performance: 168.513943899378 iterations/seconds
'''


# # ALLOW_PROFILE = True
# ALLOW_PROFILE = False

# set_profiler(ALLOW_PROFILE)


class SimpleClass:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    
    def __eq__(self, other):
        if isinstance(other, SimpleClass):
            return (self.a == other.a) and (self.b == other.b)
        
        return False


@dataclass
class DataClass:
    x: int
    y: 'Any'
    z: int = 5

    def add_one(self):
        return self.x + 1


transport: List = list()


def on_new_class(class_info, *args, **kwargs):
    transport.append(
        (
            0,
            class_info,
        )
    )


def on_new_object(adjusted_object_info, *args, **kwargs):
    transport.append(
        (
            1,
            adjusted_object_info,
        )
    )


transport_marshal: List = list()


def on_new_class_marshal(class_info, *args, **kwargs):
    transport.append(
        marshal_dumps((
            0,
            class_info,
        ))
    )


def on_new_object_marshal(adjusted_object_info, *args, **kwargs):
    transport.append(
        marshal_dumps((
            1,
            adjusted_object_info,
        ))
    )


class CoDec:
    def __init__(self, serializer_type: Serializers, single_benchmarking_item: Any, universal: bool = False, optimized: bool = True,
                 test_case: Optional[TestCase] = None) -> None:
        self.serializer: Serializer = Serializer(serializer_type)
        self.transport: List = [None] * 1024
        self.transport.clear()
        self.universal: bool = universal
        self.optimized: bool = optimized
        self.test_case: Optional[TestCase] = test_case
        if universal:
            if optimized:
                remote_objects_manager_type = UniversalOptimizedRemoteObjectsManager
            else:
                remote_objects_manager_type = UniversalRemoteObjectsManager
        else:
            if optimized:
                remote_objects_manager_type = RemoteObjectsManager
            else:
                remote_objects_manager_type = FastRemoteObjectsManager

        self.rom_src: RemoteObjectsManager = remote_objects_manager_type(
            on_new_class_handler=self.on_new_class,
            on_new_obj_info_handler=self.on_new_object,
            serializable_data_types=SUPPORTED_TYPES[serializer_type],
        )
        self.rom_dest: RemoteObjectsManager = remote_objects_manager_type(
            on_new_class_handler=self.on_new_class,
            on_new_obj_info_handler=self.on_new_object,
            serializable_data_types=SUPPORTED_TYPES[serializer_type],
        )

        if self.rom_src.serializable_tuple:
            self.delimiter = self.serializer.dumps((
                2,
                None if self.rom_src.serializable_none else 0,
            ))
        elif self.rom_src.serializable_list:
            self.delimiter = self.serializer.dumps([
                2,
                None if self.rom_src.serializable_none else 0,
            ])
        elif self.rom_src.serializable_dict:
            self.delimiter = self.serializer.dumps({
                self.rom_src.ats(DataType.int_, 0): self.rom_src.ats(DataType.int_, 2),
                self.rom_src.ats(DataType.int_, 1): None if self.rom_src.serializable_none else self.rom_src.ats(DataType.int_, 0),
            })
        else:
            raise ValueError('Serializer does not support any of the required types (tuple, list, dict).')
        
        self.single_benchmarking_item: Any = single_benchmarking_item

    def on_new_class(self, class_info, *args, **kwargs):
        if self.rom_src.serializable_tuple:
            self.transport.append(
                self.serializer.dumps((
                    0,
                    class_info,
                ))
            )
        elif self.rom_src.serializable_list:
            self.transport.append(
                self.serializer.dumps([
                    0,
                    class_info,
                ])
            )
        elif self.rom_src.serializable_dict:
            self.transport.append(
                self.serializer.dumps({
                    self.rom_src.ats(DataType.int_, 0): self.rom_src.ats(DataType.int_, 0),
                    self.rom_src.ats(DataType.int_, 1): class_info,
                })
            )
        else:
            raise ValueError('Serializer does not support any of the required types (tuple, list, dict).')

    def on_new_object(self, adjusted_object_info, *args, **kwargs):
        if self.rom_src.serializable_tuple:
            self.transport.append(
                self.serializer.dumps((
                    1,
                    adjusted_object_info,
                ))
            )
        elif self.rom_src.serializable_list:
            self.transport.append(
                self.serializer.dumps([
                    1,
                    adjusted_object_info,
                ])
            )
        elif self.rom_src.serializable_dict:
            self.transport.append(
                self.serializer.dumps({
                    self.rom_src.ats(DataType.int_, 0): self.rom_src.ats(DataType.int_, 1),
                    self.rom_src.ats(DataType.int_, 1): adjusted_object_info,
                })
            )
        else:
            raise ValueError('Serializer does not support any of the required types (tuple, list, dict).')
    
    def send_obj(self, obj) -> Tuple[bool, int]:
        result = self.rom_src.serialize(obj)
        self.transport.append(self.delimiter)
        return result
    
    def receive_objects(self) -> Generator[Tuple[int, Any], None, None]:
        try:
            for item in self.transport:
                data_info = self.serializer.loads(item)
                if self.rom_src.serializable_tuple or self.rom_src.serializable_list:
                    item_type, item_info = data_info
                elif self.rom_src.serializable_dict:
                    item_type = self.rom_src.afs(DataType.int_, data_info[self.rom_src.ats(DataType.int_, 0)])
                    item_info = data_info[self.rom_src.ats(DataType.int_, 1)]
                else:
                    raise ValueError('Serializer does not support any of the required types (tuple, list, dict).')

                if self.optimized:
                    if 0 == item_type:
                        received_class_exists, received_class_id, received_class_value = self.rom_dest.deserialize_class(item_info)
                    elif 1 == item_type:
                        received_obj_exists, received_obj_id, received_obj_value = self.rom_dest.deserialize_obj(item_info)
                    elif 2 == item_type:
                        yield received_obj_id, received_obj_value
                    else:
                        raise ValueError('Unknown item type')
                else:
                    if 0 == item_type:
                        received_class_result: ResultExistence[Tuple[int, Type]] = self.rom_dest.deserialize_class(item_info)
                        received_class_exists = bool(received_class_result)
                        received_class_id, received_class_value = received_class_result.value if received_class_exists else (None, None)
                    elif 1 == item_type:
                        received_obj_result: ResultExistence[Tuple[int, Any]] = self.rom_dest.deserialize_obj(item_info)
                        received_obj_exists = bool(received_obj_result)
                        received_obj_id, received_obj_value = received_obj_result.value if received_obj_exists else (None, None)
                    elif 2 == item_type:
                        yield received_obj_id, received_obj_value
                    else:
                        raise ValueError('Unknown item type')
        finally:
            self.transport.clear()
    
    def benchmark_single_item(self):
        self.send_obj(self.single_benchmarking_item)
        for received_obj_id, received_obj in self.receive_objects():
            if self.test_case is not None:
                self.test_case.assertEqual(self.single_benchmarking_item, received_obj)


def benchmark(test_case: TestCase):
    transport.clear()
    rom_src: RemoteObjectsManager = RemoteObjectsManager(
        on_new_class_handler=on_new_class,
        on_new_obj_info_handler=on_new_object,
    )
    rom_dest: RemoteObjectsManager = RemoteObjectsManager(
        on_new_class_handler=on_new_class,
        on_new_obj_info_handler=on_new_object,
    )

    transport_marshal.clear()
    rom_src_marshal: RemoteObjectsManager = RemoteObjectsManager(
        on_new_class_handler=on_new_class_marshal,
        on_new_obj_info_handler=on_new_object_marshal,
        serializable_data_types=SUPPORTED_TYPES[Serializers.marshal],
    )
    rom_dest_marshal: RemoteObjectsManager = RemoteObjectsManager(
        on_new_class_handler=on_new_class_marshal,
        on_new_obj_info_handler=on_new_object_marshal,
        serializable_data_types=SUPPORTED_TYPES[Serializers.marshal],
    )

    dtcl: DataClass = DataClass(5, SimpleClass(1, 2))
    dtcl.add_one()
    obj = ({'a': 1, 'b': 2}, dtcl)
    def func():
        transport.clear()
        obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        for item in transport:
            item_type, item_info = item
            if 0 == item_type:
                received_class_exists, received_class_id, received_class_value = rom_dest.deserialize_class(item_info)
            elif 1 == item_type:
                received_obj_exists, received_obj_id, received_obj_value = rom_dest.deserialize_obj(item_info)
    
        # received_object_id, received_obj = received_obj_value
        # dict_, obj_ = received_obj
        # print(dict_)
        # print(obj_)
    
    def func_marshal():
        transport_marshal.clear()
        obj_id_result: ResultExistence[int] = rom_src_marshal.serialize(obj)
        for item in transport_marshal:
            item_type, item_info = marshal_loads(item)
            if 0 == item_type:
                received_class_exists, received_class_id, received_class_value = rom_dest_marshal.deserialize_class(item_info)
            elif 1 == item_type:
                received_obj_exists, received_obj_id, received_obj_value = rom_dest_marshal.deserialize_obj(item_info)
    
        # received_object_id, received_obj = received_obj_value
        # dict_, obj_ = received_obj
        # print(dict_)
        # print(obj_)
    
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func, 0.1, 'DataClass', do_print=True)
        print()
    
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func_marshal, 0.1, 'DataClass', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.small)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func, 0.1, 'TestDataType.small', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.small)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func_marshal, 0.1, 'TestDataType.small', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.deep_small)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func, 0.1, 'TestDataType.deep_small', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.deep_small)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func_marshal, 0.1, 'TestDataType.deep_small', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.large)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func, 0.1, 'TestDataType.large', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.large)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func_marshal, 0.1, 'TestDataType.large', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.deep_large)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func, 0.1, 'TestDataType.deep_large', do_print=True)
        print()
    
    obj = test_data_factory(TestDataType.deep_large)
    with DisableGC():
    # with EnableGC():
        # transport.clear()
        # obj_id_result: ResultExistence[int] = rom_src.serialize(obj)
        measure_func_isolated_performance(func_marshal, 0.1, 'TestDataType.deep_large', do_print=True)
        print()
    
    print('=== Serializers on DataClass: ==========================================================================')
    
    with DisableGC():
        obj = ({'a': 1, 'b': 2}, dtcl)
        best_serializers, serializers_data = get_most_efficient_serializers({DataFormats.binary,
                                                    Tags.can_use_bytes,
                                                    Tags.decode_str_as_str,
                                                    Tags.decode_list_as_list,
                                                    Tags.decode_bytes_as_bytes,
                                                    Tags.deep,
                                                    Tags.can_use_custom_types,
                                                    Tags.current_platform,
                                                    # Tags.multi_platform,
                                                },
                                                obj,
                                                0.1)
        sorted_serializers = sorted(serializers_data, key=lambda x: x[1], reverse=True)
        pprint(sorted_serializers)
        print()
    
    print('=== Serializers on TestDataType: =======================================================================')

    with DisableGC():
        test_data_types = (
            TestDataType.small,
            TestDataType.deep_small,
            TestDataType.large,
            TestDataType.deep_large,
        )
    
        for test_data_type in test_data_types:
            print(f'\t== {test_data_type}: ==')
            print(f'\t\t== current_platform: ==')
            best_serializers, serializers_data = get_most_efficient_serializers({DataFormats.any,
                                                        Tags.decode_str_as_str,
                                                        Tags.decode_list_as_list,
                                                        Tags.deep,
                                                        Tags.current_platform,
                                                    },
                                                    test_data_factory(test_data_type),
                                                    0.1)
            sorted_serializers = sorted(serializers_data, key=lambda x: x[1], reverse=True)
            pprint(sorted_serializers)
            print()

            print(f'\t\t== multi_platform: ==')
            best_serializers, serializers_data = get_most_efficient_serializers({DataFormats.any,
                                                        Tags.decode_str_as_str,
                                                        Tags.decode_list_as_list,
                                                        Tags.deep,
                                                        Tags.multi_platform,
                                                    },
                                                    test_data_factory(test_data_type),
                                                    0.1)
            sorted_serializers = sorted(serializers_data, key=lambda x: x[1], reverse=True)
            pprint(sorted_serializers)
            print()
    
    print('=== RemoteObjectsManager: ==============================================================================')

    with DisableGC():
        data_set = (
            ({'a': 1, 'b': 2}, dtcl),
            test_data_factory(TestDataType.small),
            # test_data_factory(TestDataType.deep_small),
            # test_data_factory(TestDataType.large),
            # test_data_factory(TestDataType.deep_large),
        )
        best_serializers, serializers_data = get_most_efficient_serializers({DataFormats.any,
                                                    Tags.decode_str_as_str,
                                                    Tags.decode_list_as_list,
                                                    Tags.deep,
                                                    Tags.current_platform,
                                                },
                                                test_data_factory(TestDataType.deep_small),
                                                0.1)
        sorted_serializers = sorted(serializers_data, key=lambda x: x[1], reverse=True)
        for data_item_index, data_item in enumerate(data_set):
            for serializer_type, serializer_performance, serialized_data_size in sorted_serializers:
                for universal in {False, True}:
                    for optimized in {False, True}:
                        if serializer_type in {
                                Serializers.msgspec_toml,
                            }:
                            universal = True
                            codec: CoDec = CoDec(serializer_type, data_item, universal=universal, optimized=optimized, test_case=test_case)
                        else:
                            codec = CoDec(serializer_type, data_item, universal=universal, optimized=optimized, test_case=test_case)

                        measure_func_isolated_performance(codec.benchmark_single_item, 0.1, f'{data_item_index}, {serializer_type}, {universal=}, {optimized=}', do_print=True)
                        print()


class TestBenchmark(TestCase):
    def test_benchmark(self):
        raise SkipTest('This test is not ready yet')

        benchmark(self)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['Lock']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.put_coro import PutCoro
from cengal.parallel_execution.coroutines.coro_standard_services.run_coro import RunCoro
from cengal.parallel_execution.coroutines.coro_standard_services.async_event_bus import AsyncEventBus, AsyncEventBusRequest, EventID
from cengal.parallel_execution.coroutines.coro_standard_services.wait_coro import WaitCoro, WaitCoroRequest, PutSingleCoroParams, CoroutineNotFoundError, TimeoutError
from cengal.time_management.cpu_clock_cycles import perf_counter
from cengal.math.numbers import RationalNumber
from typing import Optional, Union


async def wait_for_event(i: Interface, event: EventID):
    await i(AsyncEventBus, AsyncEventBusRequest().wait(event))
    return True


class Lock:
    def __init__(self, event: EventID, timeout: Optional[RationalNumber] = None, parent: Optional['Lock'] = None) -> None:
        self._locked: bool = False
        self._event: EventID = event
        self.timeout: Optional[RationalNumber] = timeout
        self.parent: Optional['Lock'] = parent
        self.last_lock_attempt_successful: Union[None, bool] = None
    
    @property
    def locked(self):
        if self.parent is None:
            return self._locked
        else:
            return self.parent.locked
    
    @locked.setter
    def locked(self, value: bool):
        if self.parent is None:
            self._locked = value
        else:
            self.parent.locked = value
    
    @property
    def event(self):
        if self.parent is None:
            return self._event
        else:
            return self.parent.event
    
    @event.setter
    def event(self, value: EventID):
        if self.parent is None:
            self._event = value
        else:
            self.parent.event = value
    
    def lock(self, timeout: RationalNumber):
        return Lock(self.event, timeout, self)
    
    def try_lock(self):
        return Lock(self.event, 0, self)
    
    def __enter__(self):
        if self.locked:
            if 0 == self.timeout:
                self.last_lock_attempt_successful = False
                return False
            elif self.timeout is None:
                i: Interface = current_interface()
                i(RunCoro, wait_for_event, self.event)
                self.locked = True
                self.last_lock_attempt_successful = True
                return True
            else:
                i = current_interface()
                try:
                    need_to_repeat = True
                    time_spend = 0
                    while need_to_repeat and ((self.timeout is None) or (time_spend < self.timeout)):
                        start_time: float = perf_counter()
                        wait_for_event_coro_id: CoroID = i(PutCoro, wait_for_event, self.event)
                        try:
                            i(WaitCoro, WaitCoroRequest(self.timeout - time_spend, kill_on_timeout=True, result_required=True).single(wait_for_event_coro_id))
                        except CoroutineNotFoundError:
                            pass

                        time_spend += perf_counter() - start_time
                        if not self.locked:
                            self.locked = True
                            self.last_lock_attempt_successful = True
                            need_to_repeat = False
                            return True
                except TimeoutError:
                    pass
                
                return False
        else:
            self.locked = True
            self.last_lock_attempt_successful = True
            return True
    
    def __exit__(self, type, value: Exception, traceback):
        self.locked = False
        i: Interface = current_interface()
        i(AsyncEventBus, AsyncEventBusRequest().send_event(self.event, None))

    async def __aenter__(self):
        if self.locked:
            if 0 == self.timeout:
                self.last_lock_attempt_successful = False
                return False
            elif self.timeout is None:
                i: Interface = current_interface()
                await i(RunCoro, wait_for_event, self.event)
                self.locked = True
                self.last_lock_attempt_successful = True
                return True
            else:
                i = current_interface()
                try:
                    need_to_repeat = True
                    time_spend = 0
                    while need_to_repeat and ((self.timeout is None) or (time_spend < self.timeout)):
                        start_time: float = perf_counter()
                        wait_for_event_coro_id: CoroID = await i(PutCoro, wait_for_event, self.event)
                        try:
                            await i(WaitCoro, WaitCoroRequest(self.timeout - time_spend, kill_on_timeout=True, result_required=True).single(wait_for_event_coro_id))
                        except CoroutineNotFoundError:
                            pass

                        time_spend += perf_counter() - start_time
                        if not self.locked:
                            self.locked = True
                            self.last_lock_attempt_successful = True
                            need_to_repeat = False
                            return True
                except TimeoutError:
                    pass
                
                return False
        else:
            self.locked = True
            self.last_lock_attempt_successful = True
            return True

    async def __aexit__(self, type, value, traceback):
        self.locked = False
        i: Interface = current_interface()
        await i(AsyncEventBus, AsyncEventBusRequest().send_event(self.event, None))



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from time import perf_counter
from cengal.introspection.inspect.versions.v_0.inspect import get_exception
from cengal.parallel_execution.coroutines.coro_scheduler import *
from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield.versions.v_0.loop_yield import CoroPriority
from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
from cengal.parallel_execution.coroutines.coro_tools.await_coro import *
from enum import Enum
from typing import Callable, Dict, Hashable, Tuple, Union, Type, Optional, Any, List, Set
from cengal.time_management.repeat_for_a_time import Tracer
from cengal.code_flow_control.smart_values.versions.v_1 import ValueExistence
from async_generator import asynccontextmanager, async_generator, yield_
import asyncio
from tkinter import Tk, TclError
import inspect
from functools import partial
from inspect import isfunction, ismethod
from copy import copy
from cengal.time_management.sleep_tools import get_usable_min_sleep_interval


TkObjId = int


class TkinterServiceRequest(ServiceRequest):
    def create(self, tk_class: Type[Tk], *args, **kwargs) -> TkObjId:
        return self._save(0, tk_class, *args, **kwargs)

    def put(self, tk_obj: Tk) -> TkObjId:
        return self._save(1, tk_obj)

    def get(self, tk_obj_id: TkObjId) -> Tk:
        return self._save(2, tk_obj_id)

    def destroy(self, tk_obj_id: TkObjId, just_mark_as_destroyed: bool = False) -> None:
        return self._save(3, tk_obj_id, just_mark_as_destroyed)
    
    def destroy_and_wait_for_destroyed(self, tk_obj_id: TkObjId, just_mark_as_destroyed: bool = False) -> None:
        return self._save(4, tk_obj_id, just_mark_as_destroyed)
    
    def wait_for_destroyed(self, tk_obj_id: TkObjId) -> None:
        return self._save(5, tk_obj_id)
    
    def put_coro(self, tk_obj_id: TkObjId, coro_worker: AnyWorker, *args, **kwargs) -> CoroID:
        return self._save(6, tk_obj_id, coro_worker, *args, **kwargs)
    
    def register_coro(self, tk_obj_id: TkObjId, coro_id: CoroID) -> None:
        return self._save(7, tk_obj_id, coro_id)
    
    def set_update_period(self, tk_obj_id: TkObjId, period: float) -> None:
        return self._save(8, tk_obj_id, period)


class TkObjDestroyedError(Exception):
    pass


class TkObjWrapper:
    def __init__(self, interface: Interface, tk_obj_id, tk_obj, destroy_on_finish: bool = False) -> None:
        self._interface = interface
        self._tk_obj_id = tk_obj_id
        self._tk_obj = tk_obj
        self._destroyed: bool = False
        self._just_mark_as_destroyed: bool = not destroy_on_finish
        self._service: TkinterService = self._interface._loop.get_service_instance(TkinterService)
    
    @property
    def tk(self):
        if self._destroyed:
            raise TkObjDestroyedError
        else:
            return self._tk_obj
    
    @property
    def tk_id(self):
        if self._destroyed:
            raise TkObjDestroyedError
        else:
            return self._tk_obj_id
    
    def destroy(self, just_mark_as_destroyed: bool = False):
        if not self._destroyed:
            self._destroyed = True
            self._interface(TkinterService, TkinterServiceRequest().destroy(self._tk_obj_id, just_mark_as_destroyed))
    
    async def adestroy(self, just_mark_as_destroyed: bool = False):
        if not self._destroyed:
            self._destroyed = True
            await self._interface(TkinterService, TkinterServiceRequest().destroy(self._tk_obj_id, just_mark_as_destroyed))
    
    def put_coro(self, coro_worker: AnyWorker, *args, **kwargs):
        return self._interface(TkinterService, TkinterServiceRequest().put_coro(self.tk_id, coro_worker, *args, **kwargs))
    
    async def aput_coro(self, coro_worker: AnyWorker, *args, **kwargs):
        return await self._interface(TkinterService, TkinterServiceRequest().put_coro(self.tk_id, coro_worker, *args, **kwargs))
    
    def register_coro(self, coro_id: CoroID):
        return self._interface(TkinterService, TkinterServiceRequest().register_coro(self.tk_id, coro_id))
    
    async def aregister_coro(self, coro_id: CoroID):
        return await self._interface(TkinterService, TkinterServiceRequest().register_coro(self.tk_id, coro_id))
    
    @property
    def destroyed(self):
        return self._destroyed or (not self._service._get_inline(self._tk_obj_id))
    
    def __call__(self) -> Any:
        return self.tk
    
    def __enter__(self):
        return self._tk_obj
    
    def __exit__(self, type, value, traceback):
        self.destroy()

    async def __aenter__(self):
        return self._tk_obj

    async def __aexit__(self, type, value, traceback):
        await self.adestroy()
    
    def set_update_period(self, period: float):
        return self._interface(TkinterService, TkinterServiceRequest().set_update_period(self.tk_id, period))
    
    async def aset_update_period(self, period: float):
        return await self._interface(TkinterService, TkinterServiceRequest().set_update_period(self.tk_id, period))


class TkinterContextManager:
    def __init__(self, interface: Optional[Interface] = None, tk_obj_or_id: Optional[Union[Tk, TkObjId, Type[Tk]]] = None, *args, **kwargs):
        self._interface: Interface = interface or current_interface()
        self._tk_obj_or_id: Optional[Union[Tk, TkObjId]] = tk_obj_or_id
        self._args = args
        self._kwargs = kwargs
        self._tk_obj: Tk = None
        self._tk_obj_id: Optional[TkObjId] = None
        self._tk_wrapper: TkObjWrapper = None
        self._destroy_on_finish: bool = False
    
    def __enter__(self):
        if self._tk_obj_or_id is None:
            self._destroy_on_finish = True
            self._tk_obj_or_id = self._interface(TkinterService, TkinterServiceRequest().create(Tk, *self._args, **self._kwargs))
        
        if isinstance(self._tk_obj_or_id, Tk):
            self._tk_obj = self._tk_obj_or_id
            self._tk_obj_id = self._interface(TkinterService, TkinterServiceRequest().put(self._tk_obj))
        elif inspect.isclass(self._tk_obj_or_id):
            self._destroy_on_finish = True
            if not issubclass(self._tk_obj_or_id, Tk):
                self._interface._loop.logger.warning('Given class is not a subclass of Tk')
            
            self._tk_obj_id = self._interface(TkinterService, TkinterServiceRequest().create(self._tk_obj_or_id, *self._args, **self._kwargs))
            self._tk_obj = self._interface(TkinterService, TkinterServiceRequest().get(self._tk_obj_id))
        else:
            self._tk_obj_id = self._tk_obj_or_id
            self._tk_obj = self._interface(TkinterService, TkinterServiceRequest().get(self._tk_obj_id))
        
        self._tk_wrapper = TkObjWrapper(self._interface, self._tk_obj_id, self._tk_obj, self._destroy_on_finish)
        return self._tk_wrapper
    
    def __exit__(self, type, value: Exception, traceback):
        if not self._tk_wrapper.destroyed:
            self._interface(TkinterService, TkinterServiceRequest().wait_for_destroyed(self._tk_wrapper._tk_obj_id))

    async def __aenter__(self):
        if self._tk_obj_or_id is None:
            self._destroy_on_finish = True
            self._tk_obj_or_id = await self._interface(TkinterService, TkinterServiceRequest().create(Tk, *self._args, **self._kwargs))
        
        if isinstance(self._tk_obj_or_id, Tk):
            self._tk_obj = self._tk_obj_or_id
            self._tk_obj_id = await self._interface(TkinterService, TkinterServiceRequest().put(self._tk_obj))
        elif inspect.isclass(self._tk_obj_or_id):
            self._destroy_on_finish = True
            if not issubclass(self._tk_obj_or_id, Tk):
                self._interface._loop.logger.warning('Given class is not a subclass of Tk')
            
            self._tk_obj_id = await self._interface(TkinterService, TkinterServiceRequest().create(self._tk_obj_or_id, *self._args, **self._kwargs))
            self._tk_obj = await self._interface(TkinterService, TkinterServiceRequest().get(self._tk_obj_id))
        else:
            self._tk_obj_id = self._tk_obj_or_id
            self._tk_obj = await self._interface(TkinterService, TkinterServiceRequest().get(self._tk_obj_id))
        
        self._tk_wrapper = TkObjWrapper(self._interface, self._tk_obj_id, self._tk_obj, self._destroy_on_finish)
        return self._tk_wrapper

    async def __aexit__(self, type, value, traceback):
        if not self._tk_wrapper.destroyed:
            await self._interface(TkinterService, TkinterServiceRequest().wait_for_destroyed(self._tk_wrapper._tk_obj_id))


tcm = TkinterContextManager


class TkObjNotFoundError(Exception):
    pass


class TkinterService(Service, EntityStatsMixin):
    def __init__(self, loop: CoroSchedulerType):
        super(TkinterService, self).__init__(loop)

        self._request_workers = {
            0: self._on_create,
            1: self._on_put,
            2: self._on_get,
            3: self._on_destroy,
            4: self._on_destroy_and_wait_for_destroyed,
            5: self._on_wait_for_destroyed,
            6: self._on_put_coro,
            7: self._on_register_coro,
            8: self._on_set_update_period,
        }
        
        self.standard_ui_update_interval: float = 1 / 60
        self.default_update_period: float = max(get_usable_min_sleep_interval(), self.standard_ui_update_interval)
        self.update_period: float = self.default_update_period
        self.update_periods: Dict[TkObjId, float] = dict()
        self.tk_counter = Counter()
        self.tk_by_id: Dict[TkObjId, Tk] = dict()
        self.coroutines: Dict[CoroID, Set[TkObjId]] = dict()
        self.waiting_for_destroyed: Dict[TkObjId, Set[CoroID]] = dict()
        self.new_destroyed: Dict[TkObjId, bool] = dict()
        self.new_closed: Dict[TkObjId, bool] = dict()
        self.destroyed: Set[TkObjId] = set()
        self.coro_by_tk: Dict[TkObjId, CoroID] = dict()
        self.tk_users: Dict[TkObjId, Set[CoroID]] = dict()
        self.tk_after_ids: Dict[TkObjId, Set[str]] = dict()
        
        self.updater_running: bool = False
    
    def start_tk_updater(self):
        if not self.updater_running:
            self.updater_running = True
            tk_updater_coro = self._loop.put_coro(tk_updater, self)
            tk_updater_coro.is_background_coro = True

    def get_entity_stats(self, stats_level: 'EntityStatsMixin.StatsLevel' = EntityStatsMixin.StatsLevel.debug) -> Tuple[str, Dict[str, Any]]:
        return type(self).__name__, {
            'tk_counter': self.tk_counter._index + 1,
        }

    def single_task_registration_or_immediate_processing(self, request: Optional[TkinterServiceRequest]=None
                                                         ) -> ServiceProcessingResponse:
        if request is not None:
            return self.resolve_request(request)
        return True, None, None

    def full_processing_iteration(self):
        # closed
        new_closed_bak = self.new_closed
        self.new_closed = type(new_closed_bak)()
        for tk_obj_id, just_mark_as_destroyed in new_closed_bak.items():
            self.new_destroyed[tk_obj_id] = just_mark_as_destroyed
        
        # waiting_for_destroyed
        new_destroyed_bak = self.new_destroyed
        self.new_destroyed = type(self.new_destroyed)()
        for tk_obj_id, just_mark_as_destroyed in new_destroyed_bak.items():
            if tk_obj_id in self.destroyed:
                continue
                
            self.destroyed.add(tk_obj_id)
            
            tk_obj = self.tk_by_id[tk_obj_id]
            
            # if tk_obj_id in self.tk_after_ids:
            #     after_ids = self.tk_after_ids[tk_obj_id]
            #     for after_id in after_ids:
            #         tk_obj.after_cancel(after_id)
                
            #     del self.tk_after_ids[tk_obj_id]

            # cancel_all_after(tk_obj)
            
            if not just_mark_as_destroyed:
                tk_obj.destroy()
            
            if tk_obj_id in self.tk_by_id:
                del self.tk_by_id[tk_obj_id]

            if tk_obj_id in self.waiting_for_destroyed:
                for waiter_coro_id in self.waiting_for_destroyed[tk_obj_id]:
                    self.register_response(waiter_coro_id, None, None)
                
                del self.waiting_for_destroyed[tk_obj_id]
            
            if tk_obj_id in self.coro_by_tk:
                coros_tks = self.coroutines[self.coro_by_tk[tk_obj_id]]
                del self.coro_by_tk[tk_obj_id]
                if tk_obj_id in coros_tks:
                    coros_tks.remove(tk_obj_id)

            if tk_obj_id in self.tk_users:
                tk_users = self.tk_users[tk_obj_id]
                for tk_user_coro_id in tk_users:
                    # TODO: switch to an appropriate service
                    self._loop.kill_coro_by_id(tk_user_coro_id)
            
            if tk_obj_id in self.update_periods:
                del self.update_periods[tk_obj_id]
        
        # compute min update period
        if self.update_periods:
            self.update_period = min(self.update_periods.values())
        else:
            self.update_period = self.default_update_period
        
        # general
        if (not self.new_closed) and (not self.new_destroyed):
            self.make_dead()

    def in_work(self) -> bool:
        result: bool = bool(self.new_closed) or bool(self.new_destroyed) or bool(self.update_periods)
        return self.thrifty_in_work(result)
    
    def _on_create(self, tk_class: Type[Tk], *args, **kwargs) -> ServiceProcessingResponse:
        tk = tk_class(*args, **kwargs)
        tk_id: TkObjId = self.tk_counter.get()
        on_destroy = partial(self._on_create_on_destroyed, tk_id, tk)
        tk.bind("<Destroy>", on_destroy, '+')
        old_on_close = tk.protocol("WM_DELETE_WINDOW", None)
        if (not isfunction(old_on_close)) and (not ismethod(old_on_close)):
            old_on_close = None
        
        on_create_on_closed = partial(self._on_create_on_closed, tk_id, tk, old_on_close)
        tk.protocol("WM_DELETE_WINDOW", on_create_on_closed)
        self.tk_by_id[tk_id] = tk
        coro = self.current_caller_coro_info.coro
        coro_id = self.current_caller_coro_info.coro_id
        if coro_id not in self.coroutines:
            self.coroutines[coro_id] = set()
            coro.add_on_coro_del_handler(self._on_created_coro_del_handler)
        
        self.coroutines[coro_id].add(tk_id)
        self.coro_by_tk[tk_id] = coro_id
        # after_setup(tk, 1)
        after_idle_setup(tk)
        self.start_tk_updater()
        return True, tk_id, None
    
    def _on_put(self, tk_obj: Tk) -> ServiceProcessingResponse:
        tk: Tk = tk_obj
        tk_id: TkObjId = self.tk_counter.get()
        on_destroy = partial(self._on_put_on_destroyed, tk_id, tk)
        tk.bind("<Destroy>", on_destroy, '+')
        old_on_close = tk.protocol("WM_DELETE_WINDOW", None)
        if (not isfunction(old_on_close)) and (not ismethod(old_on_close)):
            old_on_close = None
        
        on_put_on_closed = partial(self._on_put_on_closed, tk_id, tk, old_on_close)
        tk.protocol("WM_DELETE_WINDOW", on_put_on_closed)
        self.tk_by_id[tk_id] = tk
        coro = self.current_caller_coro_info.coro
        coro_id = self.current_caller_coro_info.coro_id
        if coro_id not in self.coroutines:
            self.coroutines[coro_id] = set()
            coro.add_on_coro_del_handler(self._on_put_coro_del_handler)
        
        self.coroutines[coro_id].add(tk_id)
        self.coro_by_tk[tk_id] = coro_id
        # after_setup(tk, 1)
        after_idle_setup(tk)
        self.start_tk_updater()
        return True, tk_id, None
    
    def _on_get(self, tk_obj_id: TkObjId) -> ServiceProcessingResponse:
        if tk_obj_id in self.tk_by_id:
            return True, self.tk_by_id[tk_obj_id], None
        else:
            return True, None, TkObjNotFoundError()
    
    def _get_inline(self, tk_obj_id: TkObjId) -> Optional[Tk]:
        return self.tk_by_id.get(tk_obj_id, None)
    
    def _on_destroy(self, tk_obj_id: TkObjId, just_mark_as_destroyed: bool = False) -> ServiceProcessingResponse:
        self.new_destroyed[tk_obj_id] = just_mark_as_destroyed
        self.make_live()
        return True, None, None
    
    def _on_destroy_and_wait_for_destroyed(self, tk_obj_id: TkObjId, just_mark_as_destroyed: bool = False) -> ServiceProcessingResponse:
        self.new_destroyed[tk_obj_id] = just_mark_as_destroyed
        return self._on_wait_for_destroyed(tk_obj_id)
    
    def _on_wait_for_destroyed(self, tk_obj_id: TkObjId) -> ServiceProcessingResponse:
        if tk_obj_id in self.destroyed:
            return True, None, None
        
        if tk_obj_id not in self.waiting_for_destroyed:
            self.waiting_for_destroyed[tk_obj_id] = set()
        
        self.waiting_for_destroyed[tk_obj_id].add(self.current_caller_coro_info.coro_id)
        self.make_live()
        return False, None, None
    
    def _on_put_coro(self, tk_obj_id: TkObjId, coro_worker: AnyWorker, *args, **kwargs):
        exception = None
        coro_id: CoroID = None
        try:
            # TODO: switch to an appropriate service
            coro_id = self._loop.put_coro(coro_worker, *args, **kwargs).coro_id
        except:
            exception = get_exception()
        
        self._register_coro_impl(tk_obj_id, coro_id)
        return True, coro_id, exception
    
    def _on_register_coro(self, tk_obj_id: TkObjId, coro_id: CoroID):
        exception = None
        try:
            self._register_coro_impl(tk_obj_id, coro_id)
        except:
            exception = get_exception()
        
        return True, None, exception
    
    def _register_coro_impl(self, tk_obj_id: TkObjId, coro_id: CoroID):
        if tk_obj_id not in self.tk_users:
            self.tk_users[tk_obj_id] = set()
        
        self.tk_users[tk_obj_id].add(coro_id)

    def _on_set_update_period(self, tk_obj_id: TkObjId, period: float):
        self.update_periods[tk_obj_id] = period
        
        self.make_live()
        return True, None, None

    def _on_created_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        for tk_obj_id in self.coroutines[coro.coro_id]:
            self.new_destroyed[tk_obj_id] = False
        
        self.make_live()
        return True

    def _on_put_coro_del_handler(self, coro: CoroWrapperBase) -> bool:
        for tk_obj_id in self.coroutines[coro.coro_id]:
            self.new_destroyed[tk_obj_id] = True
        
        self.make_live()
        return True
    
    def _on_create_on_destroyed(self, tk_obj_id: TkObjId, tk_obj: Tk, event):
        if event.widget is tk_obj:
            if tk_obj_id not in self.destroyed:
                self.new_destroyed[tk_obj_id] = True
        
        self.make_live()
    
    def _on_put_on_destroyed(self, tk_obj_id: TkObjId, tk_obj: Tk, event):
        if event.widget is tk_obj:
            if tk_obj_id not in self.destroyed:
                self.new_destroyed[tk_obj_id] = True
        
        self.make_live()
    
    def _on_create_on_closed(self, tk_obj_id: TkObjId, tk: Tk, previous_on_close: Optional[Callable]):
        if previous_on_close is not None:
            previous_on_close()
        
        if tk_obj_id not in self.destroyed:
            self.new_closed[tk_obj_id] = False

        self.make_live()
    
    def _on_put_on_closed(self, tk_obj_id: TkObjId, tk: Tk, previous_on_close: Optional[Callable]):
        if previous_on_close is not None:
            previous_on_close()
        
        if tk_obj_id not in self.destroyed:
            self.new_closed[tk_obj_id] = False

        self.make_live()


# def tk_updater(i: Interface, tkinter_service: TkinterService):
#     from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
#     while tkinter_service.tk_by_id:
#         tk_by_id_bak = copy(tkinter_service.tk_by_id)
#         for tk_id, tk_obj in tk_by_id_bak.items():
#             if tk_id not in tkinter_service.destroyed:
#                 tk_obj.update()
        
#         i(Sleep, tkinter_service.update_period)
    
#     tkinter_service.updater_running = False
    

# def tk_updater(i: Interface, tkinter_service: TkinterService):
#     """
#     https://stackoverflow.com/questions/4083796/how-do-i-run-unittest-on-a-tkinter-app
#     https://github.com/ipython/ipython/blob/master/IPython/terminal/pt_inputhooks/tk.py

#     Args:
#         i (Interface): [description]
#         tkinter_service (TkinterService): [description]
#     """
#     from _tkinter import ALL_EVENTS as _tkinter__ALL_EVENTS, DONT_WAIT as _tkinter__DONT_WAIT
#     from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
#     from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import get_loop_yield
#     ly = get_loop_yield(CoroPriority.low)
#     while tkinter_service.tk_by_id:
#         tk_by_id_bak = copy(tkinter_service.tk_by_id)
#         for tk_id, tk_obj in tk_by_id_bak.items():
#             ly()
#             if tk_id not in tkinter_service.destroyed:
#                 while tk_obj.dooneevent(_tkinter__ALL_EVENTS | _tkinter__DONT_WAIT):
#                     ly()
        
#         i(Sleep, tkinter_service.update_period)
    
#     tkinter_service.updater_running = False
    

# def tk_updater(i: Interface, tkinter_service: TkinterService):
#     """
#     https://stackoverflow.com/questions/4083796/how-do-i-run-unittest-on-a-tkinter-app
#     https://github.com/ipython/ipython/blob/master/IPython/terminal/pt_inputhooks/tk.py

#     Args:
#         i (Interface): [description]
#         tkinter_service (TkinterService): [description]
#     """
#     from _tkinter import ALL_EVENTS as _tkinter__ALL_EVENTS, DONT_WAIT as _tkinter__DONT_WAIT
#     from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
#     from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import get_loop_yield
#     ly = get_loop_yield(CoroPriority.low)
#     while tkinter_service.tk_by_id:
#         ly()
#         tk_by_id_bak = copy(tkinter_service.tk_by_id)
#         tk_objects_with_events = set()
#         for tk_id, tk_obj in tk_by_id_bak.items():
#             if tk_id not in tkinter_service.destroyed:
#                 tk_objects_with_events.add(tk_obj)
        
#         while tk_objects_with_events:
#             new_tk_objects_with_events = set()
#             for tk_obj in tk_objects_with_events:
#                 ly()
#                 if tk_obj.dooneevent(_tkinter__ALL_EVENTS | _tkinter__DONT_WAIT):
#                     new_tk_objects_with_events.add(tk_obj)
            
#             tk_objects_with_events = new_tk_objects_with_events
        
#         i(Sleep, tkinter_service.update_period)
    
#     tkinter_service.updater_running = False


TkinterServiceRequest.default_service_type = TkinterService


# TODO: currently it uses Yield if delay is bigger that a magic number `0.001` seconds. Maybe make it some how use TkinterService.update_periods or somethin like it instead?
def tk_updater(i: Interface, tkinter_service: TkinterService):
    """
    https://stackoverflow.com/questions/4083796/how-do-i-run-unittest-on-a-tkinter-app
    https://github.com/ipython/ipython/blob/master/IPython/terminal/pt_inputhooks/tk.py

    Args:
        i (Interface): [description]
        tkinter_service (TkinterService): [description]
    """
    from _tkinter import ALL_EVENTS as _tkinter__ALL_EVENTS, WINDOW_EVENTS as _tkinter__WINDOW_EVENTS, FILE_EVENTS as _tkinter__FILE_EVENTS, TIMER_EVENTS as _tkinter__TIMER_EVENTS, IDLE_EVENTS as _tkinter__IDLE_EVENTS, DONT_WAIT as _tkinter__DONT_WAIT
    from cengal.parallel_execution.coroutines.coro_standard_services.sleep import Sleep
    from cengal.parallel_execution.coroutines.coro_standard_services.loop_yield import get_loop_yield
    ly = get_loop_yield(CoroPriority.low)
    while tkinter_service.tk_by_id:
        ly()
        tk_by_id_bak = copy(tkinter_service.tk_by_id)
        desired_length = len(tk_by_id_bak.keys())
        tk_ids_without_events = set()
        
        while len(tk_ids_without_events) < desired_length:
            for tk_id, tk_obj in tk_by_id_bak.items():
                if tk_id in tk_ids_without_events:
                    continue
                
                if tk_id in tkinter_service.destroyed:
                    tk_ids_without_events.add(tk_id)
                    continue
                
                start = perf_counter()
                has_window_events = tk_obj.dooneevent(_tkinter__WINDOW_EVENTS | _tkinter__DONT_WAIT)
                stop = perf_counter()
                if (stop - start) > 0.001:
                    i(Yield)
                # ly()
                
                start = perf_counter()
                has_file_events = tk_obj.dooneevent(_tkinter__FILE_EVENTS | _tkinter__DONT_WAIT)
                stop = perf_counter()
                if (stop - start) > 0.001:
                    i(Yield)
                # ly()
                
                start = perf_counter()
                has_timer_events = tk_obj.dooneevent(_tkinter__TIMER_EVENTS | _tkinter__DONT_WAIT)
                stop = perf_counter()
                if (stop - start) > 0.001:
                    i(Yield)
                # ly()
                
                has_events = has_window_events or has_file_events or has_timer_events
                
                if not has_events:
                    start = perf_counter()
                    tk_obj.dooneevent(_tkinter__IDLE_EVENTS | _tkinter__DONT_WAIT)
                    stop = perf_counter()
                    if (stop - start) > 0.001:
                        i(Yield)
                    # ly()
                    tk_ids_without_events.add(tk_id)

        i(Sleep, tkinter_service.update_period)
    
    tkinter_service.updater_running = False


def after_func(root):
    from cengal.parallel_execution.coroutines.coro_standard_services.simple_yield import Yield
    from cengal.parallel_execution.coroutines.coro_scheduler import current_interface, OutsideCoroSchedulerContext
    try:
        i = current_interface()
        if i is not None:
            i(Yield)
    except OutsideCoroSchedulerContext:
        pass
    
    # after_setup(root, 1)
    after_idle_setup(root)


def after_setup(root, update_period) -> bool:
    try:
        handler = root.after(update_period, after_func, root)
    except TclError:
        return False

    return True


def after_idle_setup(root) -> bool:
    try:
        handler = root.after_idle(after_setup, root, 0)
    except TclError:
        return False

    return True



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""


__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


__all__ = ['TimerHandler', 'TimeInSeconds', 'TimerRequest', 'Timer']


from cengal.time_management.cpu_clock import cpu_clock
from cengal.math.numbers import RationalNumber
from cengal.data_generation.id_generator import IDGenerator, GeneratorType
from collections import deque
from bisect import bisect_left
from typing import Any, Callable, Optional, Set, Dict, Tuple, List, Deque

TimerHandler = Callable
TimeInSeconds = RationalNumber
RequestId = int


class TimerRequest:
    def __init__(self, id: RequestId, timer_handler: TimerHandler, desired_time: TimeInSeconds):
        self.request_id: RequestId = id
        self.timer_handler = timer_handler
        self.requested_time = desired_time
        self.start_time = cpu_clock()
        self.real_end_time = None
        self.processed = False

    def __call__(self, current_time = None) -> bool:
        if not self.processed:
            if current_time is None: current_time = cpu_clock()
            if self.start_time > current_time:
                self.start_time = current_time
            time_delta = current_time - self.start_time
            if time_delta >= self.requested_time:
                self.timer_handler()
                self.processed = True
        
        return self.processed

    def time_left(self, current_time = None) -> Optional[TimeInSeconds]:
        if self.processed:
            return None

        if current_time is None: current_time = cpu_clock()
        if self.start_time > current_time:
            self.start_time = current_time
        time_delta = current_time - self.start_time
        if time_delta >= self.requested_time:
            return 0
        else:
            return self.requested_time - time_delta


class Timer:
    # TODO: (old todo) improve algorithm using `cengal\time_management\repeat_for_a_time` module
    def __init__(self):
        self.request_id_generator: IDGenerator = IDGenerator(GeneratorType.integer)
        self.requests_dict: Dict[RequestId, TimerRequest] = dict()
        self.requests_order: Deque[Tuple[RequestId, TimeInSeconds]] = deque()
        self.times_left_order: Deque[TimeInSeconds] = deque()
        self.last_time: Optional[TimeInSeconds] = None

    def register(self, timer_handler: TimerHandler, desired_time: TimeInSeconds) -> TimerRequest:
        request_id = self.request_id_generator()
        request = TimerRequest(request_id, timer_handler, desired_time)
        self.requests_dict[request_id] = request
        if self.requests_order:
            self._fast_update_requests_order(request_id, desired_time)
        else:
            self._init_requests_order(request_id, desired_time)
        
        return request
    
    def _init_requests_order(self, request_id, desired_time):
        current_time = cpu_clock()
        self.last_time = current_time
        self.requests_order.append((request_id, desired_time))
        self.times_left_order.append(desired_time)
    
    def _fast_update_requests_order(self, request_id: RequestId, desired_time: TimeInSeconds):
        current_time: TimeInSeconds = cpu_clock()
        time_left: TimeInSeconds = desired_time + (current_time - self.last_time)
        if self.requests_order[0][1] >= time_left:
            self.requests_order.appendleft((request_id, time_left))
            self.times_left_order.appendleft(time_left)
        elif self.requests_order[-1][1] <= time_left:
            self.requests_order.append((request_id, time_left))
            self.times_left_order.append(time_left)
        else:
            insert_pos: int = bisect_left(self.times_left_order, time_left)
            self.requests_order.insert(insert_pos, (request_id, time_left))
            self.times_left_order.insert(insert_pos, time_left)
    
    # Not needed anymore since much slower
    def _update_requests_order(self):
        current_time = cpu_clock()
        self.last_time = current_time
        requests_times_left = ((request.request_id, request.time_left(current_time)) for request in self.requests_dict.values())
        self.requests_order = deque((request_and_time_left for request_and_time_left in sorted(requests_times_left, key=lambda request_and_time_left: request_and_time_left[1])))
        self.times_left_order = deque((time_left for _, time_left in self.requests_order))

    def discard(self, timer_request: TimerRequest):
        try:
            del self.requests_dict[timer_request.request_id]
            return True
        except KeyError:
            return False

    def __call__(self, current_time = None):
        if current_time is None: current_time = cpu_clock()
        time_spend = current_time - self.last_time
        need_to_process_num: int = 0
        for _, time_left in self.requests_order:
            if time_left > time_spend:
                break
            
            need_to_process_num += 1
        
        if need_to_process_num:
            need_to_process: List[TimerRequest] = list()
            for _ in range(need_to_process_num):
                need_to_process.append(self.requests_dict.pop(self.requests_order.popleft()[0]))
            
            for request in need_to_process:
                request()

    def nearest_event(self, current_time = None) -> Optional[TimeInSeconds]:
        if current_time is None: current_time = cpu_clock()
        if not self.requests_order:
            return None
        
        nearest_event_time = self.requests_order[0][1] - (current_time - self.last_time)
        if nearest_event_time >= 0:
            return nearest_event_time
        else:
            return 0



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
            'PLATFORM_NAME', 'PYTHON_IMPLEMENTATION', 'PYTHON_VERSION', 'PYTHON_VERSION_STR', 
            'PYTHON_VERSION_INT', 'IS_RUNNING_IN_PYCHARM', 'RAW_OS_PLATFORM', 'OS_API_TYPE', 
            'OS_TYPE', 'KIVY_PLATFORM', 'KIVY_TARGET_PLATFORM', 'IS_RUNNING_IN_EMSCRIPTEN', 'IS_RUNNING_IN_PYODIDE', 'IS_BUILDING_FOR_PYODIDE', 
            'CENGAL_IS_IN_BUILD_MODE', 'TEMPLATE_MODULE_NAME', 'cengal_module_rel_path', 
            'cengal_module_import_str', 'current_cengal_module_rel_path', 'current_cengal_module_import_str',
            'CENGAL_UNITTESTS_DISCOVER_IS_RUNNING', 'IS_INSIDE_OR_FOR_WEB_BROWSER', 'IS_RUNNING_IN_WASI', 
            'CENGAL_VERSION_STR', 'CENGAL_VERSION_MAJOR_STR', 'CENGAL_VERSION_MINOR_STR', 
            'CENGAL_VERSION_MICRO_STR', 'CENGAL_VERSION', 'CENGAL_VERSION_MAJOR', 'CENGAL_VERSION_MINOR', 
            'CENGAL_VERSION_MICRO',
           ]


import platform
import sys
import os
from typing import Tuple, Optional


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


PLATFORM_NAME: str = platform.python_implementation()  # TODO: deprecated. Can be 'CPython', 'PyPy', 'IronPython', 'Jython'.
PYTHON_IMPLEMENTATION: str = platform.python_implementation()  # can be 'CPython', 'PyPy', 'IronPython', 'Jython'.
PYTHON_VERSION: Tuple[str, str, str] = platform.python_version_tuple()  # tuple() of str(); for example: ('3', '5', '1')
PYTHON_VERSION_STR: str = '.'.join(PYTHON_VERSION)
PYTHON_VERSION_INT: Tuple[int, int, int, str, int] = sys.version_info  # named typle sys.version_info(major=3, minor=7, micro=9, releaselevel='final', serial=0)
#   Usage: 
#       sys.version_info[0] == 3
#       (3,) > sys.version_info  # Is Python 2
#       (3, 6) <= sys.version_info  # Is Python 3.6+

IS_RUNNING_IN_PYCHARM: bool = "PYCHARM_HOSTED" in os.environ

RAW_OS_PLATFORM: str = sys.platform  # 'emscripten', 'wasi', 'darwin', 'win32', 'cygwin', 'linux', 'linux2', 'linux3', 'darwin', 'freebsd8', 'aix', aix5', 'aix7', ...
OS_API_TYPE: str = os.name  # The following names have currently been registered: 'posix', 'nt', 'java'. Android and iOS will return 'posix'.
OS_TYPE: str = platform.system()  # 'Linux', 'Windows', 'Darwin', 'Java'
try:
    from kivy.utils import platform as kivy_platform
    KIVY_PLATFORM: str = kivy_platform  # 'win', 'linux', 'android', 'macosx', 'ios', 'unknown'
except ImportError:
    KIVY_PLATFORM = 'unknown'  # 'win', 'linux', 'android', 'macosx', 'ios', 'unknown'

KIVY_TARGET_PLATFORM: str = os.environ.get('KIVY_TARGET_PLATFORM', default='unknown')  # It is required for the developer to manually set the 'KIVY_TARGET_PLATFORM' environment variable prior to the build process to ensure accurate platform detection.

IS_RUNNING_IN_EMSCRIPTEN: bool = 'emscripten' == sys.platform
IS_RUNNING_IN_WASI: bool = 'wasi' == sys.platform
IS_RUNNING_IN_PYODIDE: bool = "pyodide" in sys.modules
IS_BUILDING_FOR_PYODIDE: bool = "PYODIDE" in os.environ  # for setup.py execution time
IS_INSIDE_OR_FOR_WEB_BROWSER: bool = IS_RUNNING_IN_EMSCRIPTEN or IS_RUNNING_IN_WASI or IS_BUILDING_FOR_PYODIDE or IS_RUNNING_IN_PYODIDE

CENGAL_IS_IN_BUILD_MODE: bool = ('CENGAL_IS_IN_BUILD_MODE' in os.environ) and ('true' == os.environ['CENGAL_IS_IN_BUILD_MODE'].casefold())

TEMPLATE_MODULE_NAME: str = '_template_module'

CENGAL_UNITTESTS_DISCOVER_IS_RUNNING: bool = ('CENGAL_UNITTESTS_DISCOVER_IS_RUNNING' in os.environ) and ('true' == os.environ['CENGAL_UNITTESTS_DISCOVER_IS_RUNNING'].casefold())

CENGAL_VERSION_STR = __version__
CENGAL_VERSION = tuple([int(part) for part in CENGAL_VERSION_STR.split('.')])
CENGAL_VERSION_MAJOR = CENGAL_VERSION[0]
CENGAL_VERSION_MINOR = CENGAL_VERSION[1]
CENGAL_VERSION_MICRO = CENGAL_VERSION[2]
CENGAL_VERSION_MAJOR_STR = str(CENGAL_VERSION_MAJOR)
CENGAL_VERSION_MINOR_STR = str(CENGAL_VERSION_MINOR)
CENGAL_VERSION_MICRO_STR = str(CENGAL_VERSION_MICRO)

def cengal_module_rel_path(cengal_module) -> str:
    from cengal.modules_management.module_rel_path import module_rel_path
    import cengal
    return module_rel_path(cengal, cengal_module)


def cengal_module_import_str(cengal_module) -> str:
    from cengal.modules_management.module_rel_path import module_import_str
    import cengal
    return module_import_str(cengal, cengal_module)


def current_cengal_module_rel_path(depth: Optional[int] = 1) -> str:
    from cengal.modules_management.module_rel_path import current_module_rel_path
    import cengal
    return current_module_rel_path(cengal, depth + 1)


def current_cengal_module_import_str(depth: Optional[int] = 1) -> str:
    from cengal.modules_management.module_rel_path import current_module_import_str
    import cengal
    return current_module_import_str(cengal, depth + 1)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    'auto_gen', 
    'gather', 
    'merge_docs', 
    'gather_docs', 
]


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.file_system.directory_manager import FilteringEntity, file_list_traversal, file_list_traversal_ex, ensure_empty_dir, ensure_dir
from cengal.file_system.file_manager import last_ext
from cengal.file_system.path_manager import relative_to_src, get_relative_path_part, RelativePath
from shutil import copyfile
from os.path import join as path_join, dirname, basename
from pprint import pprint
from typing import Any, List, Set, Optional
from pathlib import Path
import shutil

from pdoc import pdoc
from pdoc import render


def auto_gen(repo_root_dir, root_dir, docs_dir, subignore_rel_path: Optional[List[str]] = None):
    subignore_rel_path = subignore_rel_path or list()
    repo_root_dir_rel = RelativePath(repo_root_dir)
    docs_dir_rel = RelativePath(docs_dir)
    src_dir_name = basename(root_dir)
    def py_filter(entity: FilteringEntity, data: Any):
        if FilteringEntity.filename == entity:
            dirpath, filename = data
            if filename in {
                '__init__.py', 
                '_template_submodule.py', 
                '__requirements__.py', 
                '_test__template_submodule.py', 
                '__x__build_config.py', 
                '__build_config.py', 
            }:
                return False
            
            if filename.startswith('test_'):
                return False

            return last_ext(filename) in {'py', 'pyw'}
        elif FilteringEntity.dirname == entity:
            dirpath, dirname = data
            if dirname in {
                '_template_module', 
                '_template_submodule', 
                'compilable', 
                'data', 
                'development', 
                'data', 
                'docs', 
                'tests', 
                'package', 
                'benchmarks', 
                '__pycache__', 
                }:
                return False

            return True
        elif FilteringEntity.dirpath == entity:
            dirpath, dirnames, filenames = data
            dirpath_basename = basename(dirpath)
            if dirpath_basename in {
                '_template_module', 
                '_template_submodule', 
                'compilable', 
                'data', 
                'development', 
                'data', 
                'docs', 
                'tests', 
                'package', 
                'benchmarks', 
                '__pycache__', 
                }:
                return False
            
            return True
        elif FilteringEntity.aggregated == entity:
            return data
        else:
            raise NotImplementedError
    
    subignore = set([repo_root_dir_rel(rel_path) for rel_path in subignore_rel_path])
    result = file_list_traversal_ex(root_dir, py_filter, True, subignore=subignore)
    gen_from_to = dict()
    gen_from = set()
    for dirpath, dirnames, filenames in result:
        for filename in filenames:
            src_path = path_join(dirpath, filename)
            gen_from.add(src_path)
    
    pprint(gen_from)
    dest_root_dir = docs_dir_rel(f'Auto-generated')
    print(dest_root_dir)
    ensure_empty_dir(dest_root_dir)
    out = Path(dest_root_dir)

    template_directory = relative_to_src()('data/pdoc-template')
    # render.configure(template_directory=template_directory)
    render.configure(docformat='markdown', search=False, template_directory=template_directory)
    for src_path in gen_from:
        try:
            print(f'Gen for <{src_path}> into {dest_root_dir}:')
            pdoc(src_path, output_directory=out)
        except:
            pass
    
    print(f'Removing unnecessary files')
    unnecessary_file_names: Set[str] = {
        'index.html',
        'search.js',
    }
    for file_name in unnecessary_file_names:
        full_file_name = out / file_name
        if full_file_name.exists():
            full_file_name.unlink()

    print(f'Renameing from "*.html" to "*.md"')
    for f in out.glob("**/*.html"):
        print(f'\tRenaming <{f}> -> "*.md"')
        f.rename(f.with_suffix(".md"))
    
    print('Done.')


def gather(repo_root_dir, root_dir, docs_dir, subignore_rel_path: Optional[List[str]] = None):
    subignore_rel_path = subignore_rel_path or list()
    repo_root_dir_rel = RelativePath(repo_root_dir)
    docs_dir_rel = RelativePath(docs_dir)
    src_dir_name = basename(root_dir)
    def md_filter(entity: FilteringEntity, data: Any):
        if FilteringEntity.filename == entity:
            dirpath, filename = data
            return 'md' == last_ext(filename)
        elif FilteringEntity.dirname == entity:
            dirpath, dirname = data
            if dirname in {
                '_template_module', 
                '_template_submodule', 
                'tests', 
                '__pycache__', 
                }:
                return False

            return True
        elif FilteringEntity.dirpath == entity:
            dirpath, dirnames, filenames = data
            dirpath_basename = basename(dirpath)
            if dirpath_basename in {
                '_template_module', 
                '_template_submodule', 
                'tests', 
                '__pycache__', 
                }:
                return False
            
            return True
        elif FilteringEntity.aggregated == entity:
            return data
        else:
            raise NotImplementedError
    
    subignore = set([repo_root_dir_rel(rel_path) for rel_path in subignore_rel_path])
    result = file_list_traversal_ex(root_dir, md_filter, True, subignore=subignore)
    copy_from_to = dict()
    for dirpath, dirnames, filenames in result:
        for filename in filenames:
            src_path = path_join(dirpath, filename)
            rel_path_part = get_relative_path_part(dirpath, root_dir)
            dest_path = path_join(src_dir_name, rel_path_part, filename)
            copy_from_to[src_path] = docs_dir_rel(dest_path)
    
    pprint(copy_from_to)
    dest_root_dir = docs_dir_rel(src_dir_name)
    print(dest_root_dir)
    ensure_empty_dir(dest_root_dir)
    for src_path, dest_path in copy_from_to.items():
        print(f'{src_path} -> {dest_path}')
        dest_file_dir = dirname(dest_path)
        print(f'{dest_file_dir}')
        ensure_dir(dest_file_dir)
        copyfile(src_path, dest_path)


def merge_docs(repo_root_dir, root_dir, docs_dir):
    docs_dir_rel = RelativePath(docs_dir)
    src_dir_name = basename(root_dir)
    docs_lib_dir = docs_dir_rel(src_dir_name)
    docs_gen_dir = docs_dir_rel('Auto-generated')
    docs_gen_dir_rel = RelativePath(docs_gen_dir)
    docs_gen_lib_dir = docs_gen_dir_rel(src_dir_name)
    def md_filter(entity: FilteringEntity, data: Any):
        if FilteringEntity.filename == entity:
            dirpath, filename = data
            return 'md' == last_ext(filename)
        elif FilteringEntity.dirname == entity:
            dirpath, dirname = data
            return True
        elif FilteringEntity.dirpath == entity:
            return True
        elif FilteringEntity.aggregated == entity:
            return data
        else:
            raise NotImplementedError
    
    result = file_list_traversal(docs_gen_lib_dir, md_filter, True)
    copy_from_to = dict()
    for dirpath, dirnames, filenames in result:
        for filename in filenames:
            src_path = path_join(dirpath, filename)
            rel_path_part = get_relative_path_part(dirpath, docs_gen_lib_dir)
            dest_path = path_join(src_dir_name, rel_path_part, filename)
            copy_from_to[src_path] = docs_dir_rel(dest_path)
    
    pprint(copy_from_to)
    dest_root_dir = docs_dir_rel(src_dir_name)
    print(dest_root_dir)
    ensure_dir(dest_root_dir)
    for src_path, dest_path in copy_from_to.items():
        print(f'{src_path} -> {dest_path}')
        dest_file_dir = dirname(dest_path)
        print(f'{dest_file_dir}')
        ensure_dir(dest_file_dir)
        copyfile(src_path, dest_path)
    
    print(f'Removing unnecessary generated docs from "{docs_gen_dir}"')
    shutil.rmtree(docs_gen_dir)
    print('Done.')


def gather_docs(repo_root_dir, root_dir, docs_dir, modules_to_ignore: Optional[List[str]] = None):
    auto_gen(repo_root_dir, root_dir, docs_dir, modules_to_ignore)
    gather(repo_root_dir, root_dir, docs_dir, modules_to_ignore)
    merge_docs(repo_root_dir, root_dir, docs_dir)


if '__main__' == __name__:
    cengal_repo_root_dir = relative_to_src()('../../../../..')
    cengal_docs_dir = RelativePath(cengal_repo_root_dir)('docs')
    cengal_src_root_dir = relative_to_src()('../../../..')
    subignore_rel_path = {
        'cengal/_examples', 
        'cengal/_template_module', 
        'cengal/code_flow_control/none_or', 
        'cengal/cross_version/console_print', 
        'cengal/help_tools', 
        'cengal/os/help_tools', 
        'cengal/parallel_execution/coroutines/coro_standard_services/cpu_tick_count_per_second', 
        'cengal/testing_lib', 
        'cengal/text_processing/docten', 
        'cengal/time_management/server_clock', 
        'cengal/universal_parser', 
        'cengal/upk_helping_tools', 
        'cengal/user_interface/console/chooser', 
    }
    auto_gen(cengal_repo_root_dir, cengal_src_root_dir, cengal_docs_dir, subignore_rel_path)
    gather(cengal_repo_root_dir, cengal_src_root_dir, cengal_docs_dir, subignore_rel_path)
    merge_docs(cengal_repo_root_dir, cengal_src_root_dir, cengal_docs_dir)



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['get_dict_item', 'get_dict_item_default', 'set_dict_item', 'change_dict_item', 'del_dict_item', 'try_del_dict_item', 'srt_to_dict_path', 'dict_path_to_str']


"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from typing import Hashable, Dict, List


def get_dict_item(data: Dict, keys: List[Hashable]):
    rv = data
    for key in keys:
        rv = rv[key]
    
    return rv


def get_dict_item_default(data: Dict, keys: List[Hashable], default=None):
    rv = data
    for key in keys:
        try:
            rv = rv[key]
        except KeyError:
            return default
    
    return rv


def set_dict_item(data: Dict, keys: List[Hashable], value):
    rv = data
    path_len = len(keys)
    for index, key in enumerate(keys):
        if index == path_len - 1:
            rv[key] = value
            break

        if key not in rv:
            rv[key] = dict()

        rv = rv[key]


def change_dict_item(data: Dict, keys: List[Hashable], value):
    rv = data
    path_len = len(keys)
    for index, key in enumerate(keys):
        if index == path_len - 1:
            rv[key] = value
            break

        rv = rv[key]
        

def del_dict_item(data: Dict, keys: List[Hashable]):
    rv = data
    path_len = len(keys)
    for index, key in enumerate(keys):
        rv = rv[key]
        if index == path_len - 2:
            del rv[keys[-1]]
            break
    
    return rv
        

def try_del_dict_item(data: Dict, keys: List[Hashable]) -> bool:
    try:
        del_dict_item(data, keys)
        return True
    except KeyError:
        return False


def srt_to_dict_path(str_path) -> List[str]:
    return eval(str_path)


def dict_path_to_str(path: List[str]) -> str:
    path_str = ', '.join([f"'{sub_path}'" for sub_path in path])
    return f'[{path_str}]'



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from enum import Enum
from time import time
from typing import Optional, Union
perf_counter = process_time = time

try:
    from cengal.time_management.cpu_clock_cycles import perf_counter
except ImportError:
    try:
        from time import perf_counter
    except ImportError:
        pass

try:
    from time import process_time
except ImportError:
    pass

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class TimeLimitIsTooSmall(Exception):
    def __init__(self, min_time: Optional[Union[float, int]], *args: object) -> None:
        super().__init__(*args)
        self.min_time: Optional[Union[float, int]] = min_time


# TODO: add support for cpu_ticks_count
class ClockType(Enum):
    fake = 0
    clock = 1
    perf_counter = 2
    process_time = 3


def _fake():
    return 0


class BaseTracer:
    """
    Base class of all tracers

    Lets assume you have a tracer:

        tr = Tracer(10.0)

    Or a fake tracer:

        tr = TracerCounter(10000, 10.0)

    As result you can use next interfaces

        tr()  # Will return True if tracer has finished it's counting and as a result, the specified time was passed.

        tr.iter_per_time_unit  # Will return counted time of iterations per second (if time() function uses second
            as a time unit)

        tr.iterations_made  # Will return number of all iterations made (not including those that were produced after 
            the end of the counting)

        tr.total_number_of_iterations_made  # Will return number of all iterations made (including those that were 
            produced after the end of the counting)

        tr.time_spent  # Will return time spent (not including time that were used after the end of the counting)

        tr.total_amount_of_time_spent  # Will return time spent (including time that were used after the end of the 
            counting)

        tr.clock_type  # Will return used time function type (ClockType enum)

    """

    def __init__(self, run_time: float, clock_type: ClockType=ClockType.perf_counter):
        self.iter = None
        self._number_of_iterations = 0
        self._last_tracked_number_of_iterations = 0
        self._start_time = None
        self._end_time = None
        self._relevant_stop_time = None
        self._last_run_was_made = False
        self._clock_type = clock_type
        self._clock = None
        self._init_clock()
        self._run_time = run_time
        if 0.0 == self._run_time:
            raise TimeLimitIsTooSmall(None)

    def __call__(self, *args, **kwargs):
        return self._last_run_was_made
    
    def _init_clock(self):
        if ClockType.fake == self._clock_type:
            self._clock = _fake
        elif ClockType.clock == self._clock_type:
            self._clock = time
        elif ClockType.perf_counter == self._clock_type:
            self._clock = perf_counter
        elif ClockType.process_time == self._clock_type:
            self._clock = process_time

    @property
    def iter_per_time_unit(self):
        raise NotImplemented()

    @property
    def iterations_made(self):
        return self._last_tracked_number_of_iterations

    @property
    def total_number_of_iterations_made(self):
        return self._number_of_iterations

    @property
    def time_spent(self):
        return self._relevant_stop_time - self._start_time

    @property
    def total_amount_of_time_spent(self):
        return self._end_time - self._start_time

    @property
    def clock_type(self) -> ClockType:
        return self._clock_type

    @clock_type.setter
    def clock_type(self, value: ClockType):
        self._clock_type = value
        self._init_clock()


class Tracer(BaseTracer):
    """
    Main tracer.
    Its task is to find out the speed of code execution, and to stop the counting at about the specified time.

    Example of use:

        tr = Tracer(10.0)
        while tr.iter():
            i = '456'
            k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iterations'.format(tr.iter_per_time_unit, tr.time_spent, tr.iterations_made))

    """

    def __init__(self, run_time: float, clock_type: ClockType=ClockType.perf_counter):
        super().__init__(run_time, clock_type)
        self._testing_worker = self._test_runs
        self._half_of_the_run_time = self._run_time / 2
        if 0.0 == self._half_of_the_run_time:
            raise TimeLimitIsTooSmall(None)
        self._relevant_start_time = None
        self._relevant_number_of_iterations_at_start = 0
        self._relevant_number_of_iterations_at_end = 0
        self._next_test_stop_on = 1
        self.iter = self._first_run

    @property
    def iter_per_time_unit(self) -> float:
        if self._last_run_was_made:
            divider = self._relevant_stop_time - self._relevant_start_time
            if 0 != divider:
                return (self._relevant_number_of_iterations_at_end - self._relevant_number_of_iterations_at_start) \
                       / divider
            return 0
        else:
            divider = self._end_time - self._start_time
            if 0 != divider:
                return self._last_tracked_number_of_iterations / divider
            return 0

    def _first_run(self) -> bool:
        self._number_of_iterations += 1
        self._relevant_start_time = self._start_time = self._clock()
        self.iter = self._subsequent_runs
        return self._test_sub_runs(self._start_time)

    def _subsequent_runs(self) -> bool:
        self._number_of_iterations += 1
        if self._number_of_iterations >= self._next_test_stop_on:
            return self._testing_worker()
        return True

    def _test_runs(self) -> bool:
        return self._test_sub_runs(self._clock())

    def _test_sub_runs(self, current_time) -> bool:
        self._relevant_stop_time = self._end_time = current_time
        self._last_tracked_number_of_iterations = self._number_of_iterations
        delta_time = current_time - self._start_time
        if delta_time >= self._half_of_the_run_time:
            time_left = self._run_time - delta_time
            # No need to:
            # if time_left < 0:
            #     time_left = 0
            iterations_per_second = self._number_of_iterations / delta_time
            iterations_left = iterations_per_second * time_left
            if iterations_left > 0:
                self._next_test_stop_on = self._number_of_iterations + round(iterations_left)
                self._testing_worker = self._last_run
                self._relevant_start_time = current_time
                self._relevant_number_of_iterations_at_start = self._number_of_iterations
                return True
            else:
                return self._last_sub_run(current_time)
        self._next_test_stop_on *= 2
        return True

    def _last_run(self) -> bool:
        return self._last_sub_run(self._clock())

    def _last_sub_run(self, current_time) -> bool:
        self._end_time = current_time
        self._last_tracked_number_of_iterations = self._number_of_iterations
        self._relevant_stop_time = self._end_time
        self._relevant_number_of_iterations_at_end = self._number_of_iterations
        self._testing_worker = self._after_last_runs
        self._last_run_was_made = True
        return False

    def _after_last_runs(self) -> bool:
        self._end_time = self._clock()
        return False


class GreedyTracer(BaseTracer):
    """
    Greedy Main tracer.
    Its task is to find out the speed of code execution, and to stop the counting at about the specified time.
    The difference is that he checks time every single iteration.

    Example of use is the same as for the Tracer()

    """

    def __init__(self, run_time: float, clock_type=ClockType.perf_counter):
        super().__init__(run_time, clock_type)
        self.iter = self._first_run

    def _first_run(self) -> bool:
        self._start_time = self._clock()
        self.iter = self._subsequent_runs
        return self._subsequent_runs()

    def _subsequent_runs(self) -> bool:
        self._relevant_stop_time = self._end_time = self._clock()
        self._number_of_iterations += 1
        self._last_tracked_number_of_iterations = self._number_of_iterations
        
        if (self._relevant_stop_time - self._start_time) < self._run_time:
            return True
        else:
            self._last_run_was_made = True
            self.iter = self._after_last_runs
            return False

    def _after_last_runs(self) -> bool:
        self._number_of_iterations += 1
        self._end_time = self._clock()
        return False

    @property
    def iter_per_time_unit(self) -> float:
        divider = self._relevant_stop_time - self._start_time
        if 0 != divider:
            return self._last_tracked_number_of_iterations / divider
        return 0


class TracerCounter(BaseTracer):
    """
    Counting tracer. Pseudo-tracer.
    Its don't have an overhead of periodic calling time() function.
    Its task is to count down within a given time, using the speed information already counted by the real tracer
    (by the Tracer class).

    Example of use:

        trc = TracerCounter(10000, 10.0)
        while trc.iter():
            i = '456'
            k = int('1243' + i)

        print('{} iter/s; {} seconds; {} iters'.format(trc.iter_per_time_unit, trc.time_spent, trc.iterations_made))

    or:
        def made_tests() -> Tracer:
            tr = Tracer(0.1)  # Run for about 0.1 of second.
            while tr.iter():
                some_my_code()
            return tr

        def run(run_time: float, tests_result: Tracer):
            trc = TracerCounter(tests_result.iter_per_time_unit, run_time)
            while trc.iter():
                some_my_code()

            print('{} iter/s; {} seconds; {} iterations'.format(trc.iter_per_time_unit, trc.time_spent,
                    trc.iterations_made))

        def main():
            tests_result = made_tests()
            ...
            while True:
                time_to_rur_str = input('Enter run time: ')
                if not time_to_run_str:
                    break
                run(float(time_to_rur_str), tests_result)

    """

    def __init__(self, iter_per_time_unit: float, run_time: float, clock_type=ClockType.fake):
        super().__init__(run_time, clock_type)
        self._iter_per_time_unit = iter_per_time_unit
        self._number_of_iterations_needed = round(self._iter_per_time_unit * self._run_time)
        self.iter = self._first_run

    def _first_run(self) -> bool:
        self._start_time = self._clock()
        self.iter = self._subsequent_runs
        return self._subsequent_runs()

    def _subsequent_runs(self) -> bool:
        self._number_of_iterations += 1
        self._last_tracked_number_of_iterations = self._number_of_iterations
        if self._number_of_iterations < self._number_of_iterations_needed:
            return True
        else:
            self._relevant_stop_time = self._end_time = self._clock()
            self._last_run_was_made = True
            self.iter = self._after_last_runs
            return False

    def _after_last_runs(self) -> bool:
        self._number_of_iterations += 1
        self._end_time = self._clock()
        return False

    @property
    def iter_per_time_unit(self) -> float:
        if self._last_run_was_made:
            divider = self._relevant_stop_time - self._start_time
            if 0 != divider:
                return self._last_tracked_number_of_iterations / divider
            return 0
        else:
            return self._iter_per_time_unit


class TracerIterator:
    """
    An iterator class. It converts any type of given tracer into an iterator.

    As result you have an option to use constructions like this:

        for i in TracerIterator(Tracer(20.0)):
            k = int('1243')

    But keep in mind that in this case there will be a bigger overhead. And there will be less CPU time for the payload.

    """

    def __init__(self, tracer: BaseTracer):
        self._tracer = tracer

    def __iter__(self):
        return self

    def __next__(self):
        if self._tracer.iter():
            return self._tracer.iterations_made
        else:
            raise StopIteration()

    next = __next__

    @property
    def tracer(self):
        return self._tracer



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cengal.modules_management.alternative_import import alt_import
from enum import Enum
UUID_PRESENT = True
with alt_import('uuid') as uuid:
    if not uuid:
        UUID_PRESENT = False

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


class GeneratorType(Enum):
    integer = 0
    guid_string = 1
    reusable_integer = 2


class IDGenerator:
    def __init__(self, generator_type=GeneratorType.integer):
        self.counter = 0
        self.type = generator_type
        self.get_new_id = None
        if GeneratorType.integer == self.type:
            self.get_new_id = self._get_new_id__int
        elif GeneratorType.guid_string == self.type:
            if UUID_PRESENT:
                self.get_new_id = self._get_new_id__uuid
            else:
                self.get_new_id = self._get_new_id__uuid_fake
        elif GeneratorType.reusable_integer == self.type:
            raise NotImplementedError
        else:
            raise RuntimeError('Wrong `generator_type` value')

    def _get_new_id__int(self):
        current_counter = self.counter
        self.counter += 1
        return current_counter

    def _get_new_id__uuid(self):
        seq = self.counter
        self.counter += 1
        current_counter = uuid.uuid1(clock_seq=seq).hex
        return current_counter

    def _get_new_id__uuid_fake(self):
        current_counter = self.counter
        self.counter += 1
        return str(current_counter)

    def remove_id(self, id_to_be_removed):
        if GeneratorType.integer == self.type:
            pass
        elif GeneratorType.guid_string == self.type:
            pass
        elif GeneratorType.reusable_integer == self.type:
            raise NotImplementedError
        else:
            raise RuntimeError('Wrong `self.type` value')

    def clear(self):
        self.counter = 0

    def __call__(self):
        return self.get_new_id()



#!/usr/bin/env python
# coding=utf-8

# Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module Docstring
Docstrings: http://www.python.org/dev/peps/pep-0257/
"""

__author__ = "ButenkoMS <gtalk@butenkoms.space>"
__copyright__ = "Copyright © 2012-2024 ButenkoMS. All rights reserved. Contacts: <gtalk@butenkoms.space>"
__credits__ = ["ButenkoMS <gtalk@butenkoms.space>", ]
__license__ = "Apache License, Version 2.0"
__version__ = "4.4.0"
__maintainer__ = "ButenkoMS <gtalk@butenkoms.space>"
__email__ = "gtalk@butenkoms.space"
# __status__ = "Prototype"
__status__ = "Development"
# __status__ = "Production"


from cengal.time_management.run_time import RT
from cengal.statistics.normal_distribution import statistics, variance as cengal_variance, standard_deviation, guess_99_95_68, count_99_95_68, average
from statistics import stdev, variance as py_variance, mean, pstdev, pvariance, median, median_low, median_high
from unittest import TestCase, main


class TestNormalDistrubution(TestCase):
    def test_compare_performance(self):
        from random import randint
        from time import time

        int_range = (0, 1000)
        sample_number = 1000000
        readings = [randint(*int_range) for _ in range(sample_number)]
        print(f'readings generated: ints in range {int_range}, {sample_number} samples')

        with RT() as rt:
            diff_mean, readings = statistics(readings)
        print(f'statistics: {rt.rt}')

        with RT() as rt:
            diff_mean, variance, max_deviation, min_deviation = cengal_variance(diff_mean, readings)
        print(f'variance: {rt.rt}')

        with RT() as rt:
            diff_mean, sd, max_deviation, min_deviation = standard_deviation(diff_mean, variance, max_deviation, min_deviation)
        print(f'standard_deviation: {rt.rt}')

        with RT() as rt:
            val_99, val_95, val_68, max_deviation, min_deviation = guess_99_95_68(diff_mean, sd, max_deviation, min_deviation)
        print(f'guess_99_95_68: {rt.rt}')

        with RT() as rt:
            val_99, val_95, val_68, max_deviation, min_deviation = count_99_95_68(readings, median)
        print(f'count_99_95_68 median: {rt.rt}')

        with RT() as rt:
            val_99, val_95, val_68, max_deviation, min_deviation = count_99_95_68(readings, average)
        print(f'count_99_95_68 average: {rt.rt}')

        with RT() as rt:
            val_99, val_95, val_68, max_deviation, min_deviation = count_99_95_68(readings, min)
        print(f'count_99_95_68 min: {rt.rt}')

        with RT() as rt:
            val_99, val_95, val_68, max_deviation, min_deviation = count_99_95_68(readings, max)
        print(f'count_99_95_68 max: {rt.rt}')

        with RT() as rt:
            average_value = average(readings)
        print(f'average: {rt.rt}')

        with RT() as rt:
            py_mean = mean(readings)
        print(f'mean: {rt.rt}')

        with RT() as rt:
            result = median(readings)
        print(f'median: {rt.rt}')

        with RT() as rt:
            result = median_low(readings)
        print(f'median_low: {rt.rt}')

        with RT() as rt:
            result = median_high(readings)
        print(f'median_high: {rt.rt}')

        with RT() as rt:
            result = py_variance(readings)
        print(f'variance: {rt.rt}')

        with RT() as rt:
            result = stdev(readings)
        print(f'stdev: {rt.rt}')

        with RT() as rt:
            result = pstdev(readings)
        print(f'pstdev: {rt.rt}')

        with RT() as rt:
            result = py_variance(readings, py_mean)
        print(f'variance with mean: {rt.rt}')

        with RT() as rt:
            result = stdev(readings, py_mean)
        print(f'stdev with mean: {rt.rt}')

        with RT() as rt:
            result = pstdev(readings, py_mean)
        print(f'pstdev with mean: {rt.rt}')


if __name__ == '__main__':
    main()
